# -*- coding: utf-8 -*-

"""

  HTTP semantic logic tests
  ~~~~~~~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# testing
from canteen import test
from canteen import base

# HTTP layer
from canteen.logic import http

# werkzeug
from werkzeug import routing, wrappers
from werkzeug.test import EnvironBuilder


class URLUtilTests(test.FrameworkTest):

  """ Tests builtin framework decorator for binding URLs """

  def _mock_handler(self):

    """ Generate a small mock object that can be introspected.

        :returns: An object. """

    return object()

  def test_url_basic(self):

    """ Test basic functionality of `http.url` decorator """

    sent = wrap = self._mock_handler()
    out = http.url(r'/sup')(wrap)
    assert out is sent is wrap

  def test_url_named(self):

    """ Test creating a named `http.url` binding """

    sent = wrap = self._mock_handler()
    out = http.url('test', r'/sup')(wrap)
    assert out is sent is wrap

  def test_url_param(self):

    """ Test creating a bound `http.url` with params in the URL """

    sent = wrap = self._mock_handler()
    out = http.url('test', r'/sup/<int:num>')(wrap)
    assert out is sent is wrap

  def test_url_wrap(self):

    """ Test the `http.url` decorator for `wrap` keyword usage """

    sent = wrap = self._mock_handler()

    def other_wrap(target):
      assert target is sent is wrap
      return target

    out = http.url('test', r'/sup', wrap=other_wrap)(wrap)
    assert out is sent is wrap


class HTTPSemanticsTests(test.FrameworkTest):

  """ Tests builtin framework logic related to HTTP semantics. """

  def _mock_environ(self, *args, **kwargs):

    """ Make a small mock WSGI environment. """

    return EnvironBuilder(*args, **kwargs).get_environ()

  def test_construct(self):

    """ Test constructing an instance of `HTTPSemantics` """

    h = http.HTTPSemantics()
    return h

  def test_injection(self):

    """ Test injection of `HTTPSemantics` on DI consumers """

    b = base.Handler({}, lambda: None)
    assert b.http

  def test_http_request(self):

    """ Test constructing an instance of `HTTPRequest` """

    r = http.HTTPSemantics.new_request(self._mock_environ())
    assert r
    assert isinstance(r, http.HTTPSemantics.HTTPRequest)
    return r

  def test_http_request_set_session(self):

    """ Test setting the current session by force on an `HTTPRequest` """

    session = {'user': 123}
    request = self.test_http_request()
    request.set_session(session)

    assert isinstance(request.session, tuple)
    assert len(request.session) == 2

  def test_http_response(self):

    """ Test constructing an instance of `HTTPResponse` """

    r = http.HTTPSemantics.new_response()
    assert r
    assert isinstance(r, http.HTTPSemantics.HTTPResponse)
    return r

  def test_add_route(self):

    """ Test force-adding a bound URL via `HTTPSemantics` """

    http.HTTPSemantics.add_route((r'/never-ever', 'testing'), lambda: 'hi')

  def test_resolve_route(self):

    """ Test resolving a route by name via `HTTPSemantics` """

    self.test_add_route()
    r = http.HTTPSemantics.resolve_route('testing')
    assert r

    r = http.HTTPSemantics.resolve_route('idonotexistnotever')
    assert not r

  def test_routes_iter(self):

    """ Test iterating through known URL routes via `HTTPSemantics` """

    for route in http.HTTPSemantics.routes:
      assert route
      assert isinstance(route, routing.Rule)

  def test_route_map(self):

    """ Test generating a route map via `HTTPSemantics` """

    assert http.HTTPSemantics.route_map
    assert isinstance(http.HTTPSemantics.route_map, routing.Map)

  def test_http_error(self):

    """ Test raising an HTTP error via `HTTPSemantics` """

    with self.assertRaises(http.HTTPSemantics.HTTPException):
      http.HTTPSemantics.error(404)

    with self.assertRaises(http.HTTPSemantics.HTTPException):
      http.HTTPSemantics.error(500)

  def test_http_redirect(self):

    """ Test preparing an HTTP redirect via `HTTPSemantics` """

    r = http.HTTPSemantics.redirect('/sample')
    assert isinstance(r, wrappers.Response)



# -*- coding: utf-8 -*-

"""

  config utils
  ~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import os
import importlib


## Globals
_appconfig = {}


class Config(object):

  """  """

  __dev__ = None  # debug mode

  seen = set()  # seen config items
  wrap = None  # wrapped config block
  blocks = None  # wrapped config blocks

  ## -- Internals -- ##
  def __init__(self, sub=None, **blocks):

    """  """

    global _appconfig

    self.blocks = blocks or _appconfig
    if not _appconfig:
      _appconfig = blocks
    self.wrap = sub

  ### === Public Attributes === ###

  # block shortcuts
  app = property(lambda self: self.blocks.get('app', {'debug', True}))
  assets = property(lambda self: self.blocks.get('assets', {'debug': True}))
  config = property(lambda self: self.blocks.get('config', {}))

  debug = lambda self: (
    any((self.__dev__,
         os.environ.get('SERVER_SOFTWARE', 'Not Dev').startswith('Dev'),
         os.environ.get('CANTEEN_DEBUG', False) in (
          '1', 'yes', 'on', 'true', 'sure', 'whynot'),
         self.config.get('debug', False),
         self.app.get('debug', False),
         __debug__)))

  app_version = lambda self: (
    '0.0.1-alpha' if 'version' not in self.app else (
      '-'.join(('.'.join(map(str, self.app['version'].values()[:3])),
                             str(self.app['version']['release'])))))

  ### === Public Methods === ###
  load = lambda self, path: (
    self.merge(importlib.import_module(path).config.blocks))
  __get__ = lambda self, instance, owner: self.wrap or self.blocks

  def get(self, key, default=None):

    """  """

    if self.blocks:  # pragma: no cover
      if 'config' in self.blocks:
        return self.blocks['config'].get(key, {'debug': True})
      return self.blocks.get(key, default)
    return default



# -*- coding: utf-8 -*-

"""

  HTTP logic tests
  ~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  __all__ = ('test_agent',
             'test_caching',
             'test_cookies',
             'test_etags',
             'test_redirects',
             'test_semantics')



# -*- coding: utf-8 -*-

"""

  test runner
  ~~~~~~~~~~~

  discovers canteen's tests, then runs them.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""


if __debug__:  # pragma: no cover

  # stdlib & canteen
  import sys, os


  if __name__ == '__main__' and (
        'CANTEEN_TESTING' not in os.environ):  # pragma: no cover
    import nose
    sys.exit(int(nose.run()))



# -*- coding: utf-8 -*-

"""

  template logic
  ~~~~~~~~~~~~~~

  exposes base logic for dealing with ``Jinja2`` templates.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

from __future__ import print_function

# stdlib
import os
import re
import sys
import json
import errno
import operator
import importlib
import py_compile

# os
from os import path, listdir

# base
from ..base import (logic,
                    handler)

# core & utils
from ..core import runtime
from ..util import decorators


## Constants
_DEFAULT_MODULE_PREFIX = 'templates'
_DEFAULT_TEMPLATE_PATTERN = re.compile(
  r'^.*\.(html|js|haml|svg|css|sass|less|scss|coffee)$')
_FRAMEWORK_ROOT_PATH = (
  path.dirname(  # /canteen
    path.dirname(__file__)))  # /canteen/logic
_FRAMEWORK_TEMPLATE_ROOT = (
  path.join(_FRAMEWORK_ROOT_PATH, 'templates'))
_FRAMEWORK_TEMPLATE_SOURCES = (
  path.join(_FRAMEWORK_TEMPLATE_ROOT, 'source'))
_FRAMEWORK_TEMPLATES_COMPILED = (
  path.join(_FRAMEWORK_TEMPLATE_ROOT, 'compiled'))

## Globals
_conditionals = []

## Utility Lambdas
average = lambda x: reduce(operator.add, x) / len(x)
_dropext = lambda _path: path.splitext(_path)[0]
_make_module = lambda mod, prefix=_DEFAULT_MODULE_PREFIX: (
  '.'.join((prefix, mod)) if prefix else mod)
_make_import_statement = lambda module: 'from %s import *' % module
_make_module_path = lambda _path, root: (
  '/'.join(filter(bool, _dropext(_path).replace(root, '')
                                       .replace('.py', '')
                                       .split('/'))).replace('/', '.'))


with runtime.Library('jinja2', strict=True) as (library, jinja2):

  # jinja internals
  from jinja2 import nodes, compiler
  from jinja2.nodes import EvalContext
  from jinja2._compat import iteritems
  from jinja2.compiler import Frame, find_undeclared
  from jinja2.compiler import unoptimize_before_dead_code
  from jinja2.ext import with_, autoescape, do, loopcontrols

  DEFAULT_EXTENSIONS = [with_, autoescape, do, loopcontrols]


  def template_ast(self, node, frame=None):  # pragma: no cover

    """ Shim for Jinja2's default ``Jinja``-sytnax-to-Python AST converter.
        Wraps template code in a module-level ``run`` function that binds it
        to an instance of :py:class:`jinja2.Environment`.

        :param node: Current AST node.
        :param frame: Current code frame.
        :return: ``None``. """

    assert frame is None, 'no root frame allowed'
    eval_ctx = EvalContext(self.environment, self.name)

    from jinja2.runtime import __all__ as exported
    self.writeline('# -*- coding: utf-8 -*-')
    self.writeline('')
    self.writeline('from __future__ import division')
    self.writeline('from jinja2.runtime import ' + ', '.join(exported))
    if not unoptimize_before_dead_code:
        self.writeline('dummy = lambda *x: None')

    # if we want a deferred initialization we cannot move the
    # environment into a local name
    envenv = not self.defer_init and ', environment=environment' or ''

    # do we have an extends tag at all?  If not, we can save some
    # overhead by just not processing any inheritance code.
    have_extends = node.find(nodes.Extends) is not None

    # find all blocks
    for block in node.find_all(nodes.Block):
        if block.name in self.blocks:
            self.fail('block %r defined twice' % block.name, block.lineno)
        self.blocks[block.name] = block

    # find all imports and import them
    for import_ in node.find_all(nodes.ImportedName):
        if import_.importname not in self.import_aliases:
            imp = import_.importname
            self.import_aliases[imp] = alias = self.temporary_identifier()
            if '.' in imp:
                module, obj = imp.rsplit('.', 1)
                self.writeline('from %s import %s as %s' %
                               (module, obj, alias))
            else:
                self.writeline('import %s as %s' % (imp, alias))

    # add the load name
    self.writeline('name = %r' % self.name)

    # generate the deferred init wrapper
    self.writeline('def run(environment):', extra=1)
    self.indent()

    # generate the root render function.
    self.writeline('def root(context%s):' % envenv, extra=1)

    # process the root
    frame = Frame(eval_ctx)
    frame.inspect(node.body)
    frame.toplevel = frame.rootlevel = True
    frame.require_output_check = have_extends and not self.has_known_extends
    self.indent()
    if have_extends:
        self.writeline('parent_template = None')
    if 'self' in find_undeclared(node.body, ('self',)):
        frame.identifiers.add_special('self')
        self.writeline('l_self = TemplateReference(context)')
    self.pull_locals(frame)
    self.pull_dependencies(node.body)
    self.blockvisit(node.body, frame)
    self.outdent()

    # make sure that the parent root is called.
    if have_extends:
        if not self.has_known_extends:
            self.indent()
            self.writeline('if parent_template is not None:')
        self.indent()
        self.writeline('for event in parent_template.'
                       'root_render_func(context):')
        self.indent()
        self.writeline('yield event')
        self.outdent(2 + (not self.has_known_extends))

    # at this point we now have the blocks collected and can visit them too.
    for name, block in iteritems(self.blocks):
        block_frame = Frame(eval_ctx)
        block_frame.inspect(block.body)
        block_frame.block = name
        self.writeline('def block_%s(context%s):' % (name, envenv),
                       block, 1)
        self.indent()
        undeclared = find_undeclared(block.body, ('self', 'super'))
        if 'self' in undeclared:
            block_frame.identifiers.add_special('self')
            self.writeline('l_self = TemplateReference(context)')
        if 'super' in undeclared:
            block_frame.identifiers.add_special('super')
            self.writeline('l_super = context.super(%r, '
                           'block_%s)' % (name, name))
        self.pull_locals(block_frame)
        self.pull_dependencies(block.body)
        self.blockvisit(block.body, block_frame)
        self.outdent()

    self.writeline('blocks = {%s}' % ', '.join('%r: block_%s' % (x, x)
                                               for x in self.blocks),
                   extra=1)

    # add a function that returns the debug info
    self.writeline('debug_info = %r' % '&'.join('%s=%s' % x for x
                                                in self.debug_info))

    self.writeline('return (root, blocks, debug_info)')
    self.outdent()
    self.writeline('')

  compiler.CodeGenerator.visit_Template = template_ast


  class TemplateCompiler(object):

    """ Compiles ``Jinja2``-format templates into raw Python for faster
        execution, better optimization and more effective caching of template
        routines. """

    def __init__(self, module, sources, target, config,
                                prefix=_DEFAULT_MODULE_PREFIX,
                                debug=False):

      """ Initialize this ``TemplateCompiler``.

          :param module: Path to folder due to be used as the root for a new
            Python package containing both (recursively) ``compiled`` templates
            and ``sources``.

          :param sources: Path to folder of template source files, due to be
            (recursively) compiled into raw Python.

          :param target: Path to folder that should be filled with new Python
            files containing compiled template logic, built recursively from
            corresponding paths in ``sources``.

          :param config: Reference to application (or empty framework-level)
            configuration.

          :param prefix: Module prefix to prepend to compiled template
            imports.

          :param debug: Debug flag (``bool``) specifying whether we should
            output a bunch of noise about what's going on. Defaults to
            ``False``, in which case nothing is outputted. """

      self.module, self.sources, self.target, self.config, self.prefix = (
        module, sources, target, config, prefix)

      self.debug = debug

      # apply again
      compiler.CodeGenerator.visit_Template = template_ast

    @property
    def environment(self):

      """ Load a clean, throwaway ``Jinja2`` :py:class:`jinja2.Environment`
          instance, for the purpose of preparing compiled templates.

          :returns: Prepped instance of :py:class:`jinja2.Environment`. """

      return Templates().environment(handler.Handler(), self.config)

    @staticmethod
    def mkdir_p(fullpath):

      """ Recursively bring a directory path into existence. Similar to the Unix
          command ``mkdir -p``.

          :param fullpath: Path to recursively make directories for.

          :raises OSError: In the event of an unexpected error (such as a
            permissions or low-level storage error).

          :returns: Nothing. """

      try:
        os.makedirs(fullpath)
      except OSError as e:
        if e.errno == errno.EEXIST and path.isdir(fullpath):
          pass
        else:  # pragma: no cover
          raise

    @classmethod
    def framework(cls, config=None, run=False):

      """ Class-level shortcut to spawn a ``TemplateCompiler`` instance that is
          pre-configured to build builtin framework templates.

          :param config: Instance of :py:class:`canteen.util.config.Config`
            that should be used when compiling templates.

          :param run: ``bool`` flag indicating that the compiler should be run
            automatically, instead of returned. Defaults to ``False``.

          :returns: Instance of ``TemplateCompiler``. """

      from canteen.util import config as configutil

      # create `compiled` root folder, if necessary
      if not path.isdir(_FRAMEWORK_TEMPLATES_COMPILED):  # pragma: no cover
        cls.mkdir_p(_FRAMEWORK_TEMPLATES_COMPILED)

      compiled = cls(_FRAMEWORK_TEMPLATE_ROOT,
                     _FRAMEWORK_TEMPLATE_SOURCES,
                     _FRAMEWORK_TEMPLATES_COMPILED,
                     config or configutil.Config(),
                     'canteen.templates')
      if run: return compiled()
      return compiled

    def compile(self, environment, source, destination,
                                              encoding='utf-8',
                                              base_dir=''):

      """ Compile an individual ``Jinja2``-formatted template into raw Python,
          using a slightly customized version of the standard Jinja template
          compiler.

          :param environment: Instance of :py:class:`jinja2.Environment` to use
            for compiling the actual template sources.

          :param source: Path to source directory that is due to be recursively
            compiled into raw Python at ``destination``.

          :param destination: Destination directory that is due to be filled
            with compiled raw Python templates.

          :param encoding: Default encoding when reading template ``source``
            files and writing template ``destination`` files.

          :param base_dir: Base directory path to be removed from the beginning
            of compiled template names.

          :raises ValueError: If a template name collides with a folder name
            and so cannot be used in a compiled fashion, since they would
            generate ambiguous import paths in Python.

          :returns: ``list`` of new template import paths that should work if
            ``destination`` is in the current interpreter's ``sys.path``. """

      with open(source) as template:

        name = source.replace(base_dir, '')  # remove base path from name

        try:
          raw = environment.compile(template.read().decode(encoding), **{
            'name': name,
            'filename': name,
            'raw': True,
            'defer_init': True})

        except jinja2.TemplateSyntaxError:  # pragma: no cover
          print("!!! Syntax error in file '%s'. Compilation failed. !!!" % (
            str(source)))

        target_name, ext = path.splitext(destination)
        if path.isdir(target_name):  # pragma: no cover
          raise ValueError('Template name conflict: `%s` is a directory,'
                           ' so %s%s cannot exist as an independent'
                           ' template file.' % (name, name, ext))

        self.mkdir_p((path.join(*tuple(destination.split('/')[0:-1]))))

        with open('.'.join((target_name, 'py')), 'w') as target:
          target.write(raw)
      return destination

    def compile_dir(self, source, destination,
                          base_dir='',
                          pattern=_DEFAULT_TEMPLATE_PATTERN,
                          encoding='utf-8',
                          fill_init=True):

      """ Recursively compile a directory of ``Jinja2``-formatted templates to
          raw Python source code.

          :param source: Source directory to compile from.

          :param destination: Destination directory to write compiled Python
            template files to.

          :param base_dir: Base directory that contains both ``source`` and
            ``destination``, if any. Manages canteen-style ``__init__.py``
            file.

          :param pattern: String regex pattern to scan files against in the
            ``source`` directory. Matching filenames are considered to be
            ``Jinja2``-format templates and will be marked for compilation.

          :param encoding: Default encoding when reading template ``source``
            files and writing template ``destination`` files.

          :param fill_init: Fill out an proper ``__init__.py``, aggressively
            importing submodules.

          :returns: Full (recursively-built) list of import paths generated
            during the compilation routine. """

      if not destination.replace(base_dir, '') == '':
        destination = path.join(base_dir, '/'.join(
          destination.replace(base_dir, '').replace('-', '_').split('/')[1:]))

      _import_paths = []

      if path.isdir(destination):  # pragma: no cover
        init = path.join(destination, '__init__.py')
        if not path.exists(init):
          open(init, 'w').close()

      for filename in listdir(source):
        source_name, destination_name = (
          path.join(source, filename),
          path.join(destination, filename.replace('-', '_')))

        # @TODO(sgammon): compiling file
        if self.debug:  # pragma: no cover
          print('Compiling %s...' % filename)

        if path.isdir(source_name):
          self.mkdir_p(destination_name)
          _import_paths += self.compile_dir(
              source_name,
              destination_name,
              base_dir,
              encoding=encoding)

        elif path.isfile(source_name) and (
            re.match(pattern, filename)):  # pragma: no cover

          # make sure directories are there
          if not path.isdir(path.dirname(source_name)):
            self.mkdir_p(path.dirname(source_name))

          new_target = _make_module_path(
            self.compile(self.environment, source_name, destination_name,
                          encoding=encoding,
                          base_dir=base_dir), base_dir)

          _import_paths.append(_make_module(new_target, self.prefix))

          file_path, file_name = tuple(destination_name.rsplit('/', 1))
          source_path = path.join(file_path, '.'.join((
            file_name.rsplit('.', 1)[0], 'py')))

          precompiled_path = path.join(file_path, '.'.join((
            file_name.rsplit('.', 1)[0], 'pyc' if (
              sys.flags.optimize) else 'pyo')))

          try:
            py_compile.compile(source_path, precompiled_path)
          except Exception:
            print("Failed to precompile: '%s'... Skipping..." % source_path)

        elif path.isfile(source_name) and not (
            re.match(pattern, filename)):  # pragma: no cover
          with open(destination_name, 'wb') as staticwrite:
            with open(source_name, 'rb') as staticread:
              staticwrite.write(staticread.read())

      if path.isdir(destination) and fill_init:
        with open(path.join(destination, '__init__.py'), 'w') as init:
          map(lambda line: init.write(line + "\n"), (
            '# -*- coding: utf-8 -*-',
            '',
            '"""',
            '',
            '   compiled templates: %s' % _make_module_path(*(
                                                      destination, base_dir)),
            '',
            '"""',
            '',
            '# subtemplates',
            "\n".join(map(_make_import_statement, _import_paths)),
            ''))

      return _import_paths

    def __call__(self, **kw):

      """ Defer calls on ``TemplateCompiler`` objects to :py:meth:`compile_dir`.

          :param kw: Keyword arguments to pass to ``compile_dir``.
          :returns: Whatever ``compile_dir`` decides to return. """

      return self.compile_dir(*(self.sources, self.target, self.module), **kw)


  class ModuleLoader(jinja2.ModuleLoader):

    """ ``Jinja2`` :py:class:`TemplateLoader`, responsible for resolving Python-
        optimized template routines. When ``Jinja2`` templates are compiled to
        raw Python, this is the loader that is responsible for
        importing/executing/preparing responses from those templates. """

    cache = None  # cache of modules loaded
    module = None  # main module
    has_source_access = False  # from jinja's internals

    def __init__(self, module='templates', strict=False):

      """ Initialize this ``ModuleLoader`` with a target top-level ``module``.
          Templates will be loaded under this top level module name, where
          directories are converted to packages and the template file has
          been converted to a Python module.

          Canteen extends *Jinja2's* builtin ``ModuleLoader`` to work with a
          slightly-customized form of compiled templates, where the compiled
          template content is wrapped in a module-level callable that binds it
          to the current ``environment``. This specialized compiler is shipped
          with Canteen at :py:mod:`canteen.logic.template.Compiler`.

          :param module: Top-level Python package to load from. Essentially
            treated as a module prefix, added before any template paths are
            converted to dotted Python module paths. Thus, a value of
            ``templates`` (the default) paired with a later template source
            request for `pages/about.html` would result in an import call for
            the module path ``templates.pages.about``.

          :param strict: Whether to silently or loudly fail if the given
            ``module`` could not be loaded. ``bool`` that defaults to ``False``,
            meaning we should fail silently and later fail all template requests
            with a ``jinja2.TemplateNotFound`` (until the import succeeds, as it
            will be retried each time). ``True`` will raise an ``ImportError``
            from this function if the module is invalid.

          :raises ImportError: If ``strict`` is set to ``True`` and the given
            ``module`` is invalid or could not be loaded. """

      from canteen.logic.cache import Caching

      if isinstance(module, basestring):
        try:
          module = importlib.import_module(module)
        except ImportError:
          if strict: raise

      self.cache, self.module = (
        Caching.spawn('tpl_%s' % module if (
          isinstance(module, basestring)) else module.__name__),
        module)

    def load(self, environment, filename, _globals=None):

      """ Load a compiled ``Jinja2`` template from an ``environment`` and
          template ``filename``.

          :param environment: Currently-active ``Jinja2`` template execution
            environment. Instance of :py:class:`jinja2.Environment`.

          :param filename: Filename for the final template to be served.
            Resolved to a Python module and executed to prepare the template
            routines.

          :param _globals: Global-level ``key=>value`` pairs (in a ``dict``) to
            be applied to the local rendering ``environment`` before render.

          :returns: Loaded :py:class:`jinja2.Template` object, an instance of
            the current ``environment``'s :py:attr:`template_class`. Prepared
            with bound ``_globals``, ``environment`` and ``blocks``. """

      _globals = _globals or {}

      if isinstance(self.module, basestring):  # pragma: no cover
        self.module = importlib.import_module(self.module)

      filename, ext = path.splitext(filename.strip('/'))

      t = self.cache.get(filename)
      if not t:
        # Store module to avoid unnecessary repeated imports.
        mod = self.get_module(filename)

        # initialize module
        root, blocks, debug_info = mod.run(environment)

        # manufacture new template object from cached module
        t = object.__new__(environment.template_class)
        t.environment, t.globals, t.name, t.filename, t.blocks  = (
          environment,
          _globals,
          mod.name,
          filename,
          blocks)

        # jinja2 internals
        t.root_render_func, t._module, t._debug_info, t._uptodate = (
          root,
          None,
          debug_info,
          lambda: True)

        self.cache.set(filename, t)
      return t

    def get_module(self, template):

      """ Converts a template path to a package path and attempts import, or
          else raises ``Jinja2``'s :py:class:`TemplateNotFound`.

          :param template: Full string filepath to the ``template`` desired.
            Converted to a Python-style dotted-path for import.

          :raises jinja2.TemplateNotFound: If the template in question could not
            be located or loaded properly in the current environment.

          :returns: Template module in question. """

      import jinja2

      fullpath = self.module.__name__.split('.') + template.split('/')
      prefix, obj = '.'.join(fullpath[:-1]), fullpath[-1]

      try:
        return getattr(__import__(prefix, None, None, [obj]), obj)
      except (ImportError, AttributeError):
        raise jinja2.TemplateNotFound(template)


  class FileLoader(jinja2.FileSystemLoader):

    """ ``Jinja2`` template :py:class:`jinja2.Loader` subclass that loads
        template source files and indicates access to template sources. """

    has_source_access = True


@decorators.bind('template', namespace=False)
class Templates(logic.Logic):

  """ Provides logic related to templating for Canteen and Canteen-based apps.
      Currently supports ``Jinja2``, optionally with ``HAMLish``-flavored
      template syntax. """

  with runtime.Library('jinja2') as (library, jinja2):

    """ Prepares Jinja2 for integration with Canteen. """

    ## == Attributes == ##
    engine = jinja2  # we're using jinja :)
    default_extensions = property(lambda self: DEFAULT_EXTENSIONS)
    default_config = property(lambda self: {
      'optimized': True, 'autoescape': True})

    @staticmethod  # default syntax support method
    def _default_syntax(handler, environment_factory, j2config, config):

      """ Prepare the default template syntax, which is just ``Jinja2``'s
          regular template syntax.

          :param handler: Currently-active web handler. Always an instance of
            :py:class:`canteen.base.handler.Handler` or a subtype thereof.

          :param environment_factory: Callable that is charged with producing
            ``Jinja2`` :py:class:`jinja2.Environment` objects.

          :param j2config: ``Jinja2``-specific configuration, extracted from
            app config. These are *kwarg* parameters in ``Jinja2``'s docs for
            :py:class:`jinja2.Environment`.

          :param config: Reference to the full application config object, always
            an instance of :py:class:`canteen.util.config.Config`.

          :returns: Instance of :py:class:`jinja2.Environment`, prepared with
            ``Jinja2``'s default template syntax. """

      # factory environment
      return environment_factory(**j2config)

    syntax = _default_syntax

    # is there HAML syntax support?
    with runtime.Library('hamlish_jinja') as (haml_library, haml):

      """ Prepares HAMLish Jinja for integration with Canteen. """

      # we're using haml :)
      syntax_extension = (haml.HamlishExtension, haml.HamlishTagExtension)

      def _hamlish_syntax(self, handler, environment_factory, j2config, config):

        """ Prepare a ``HAMLish``-based tempmlate syntax, which is based on
            Ruby's *HAML*, but is extended/modified to be a bit more Pythonic
            and friendly to ``Jinja2``.

            :param handler: Currently-active web handler. Always an instance of
              :py:class:`canteen.base.handler.Handler` or a subtype thereof.

            :param environment_factory: Callable that is charged with producing
              ``Jinja2`` :py:class:`jinja2.Environment` objects.

            :param j2config: ``Jinja2``-specific configuration, extracted from
              app config. These are *kwarg* parameters in ``Jinja2``'s docs for
              :py:class:`jinja2.Environment`.

            :param config: Reference to the full application config object,
              always an instance of :py:class:`canteen.util.config.Config`.

            :returns: Instance of :py:class:`jinja2.Environment`, prepared with
              ``HAMLish``-based ``Jinja2`` template syntax. """

        # make environment first
        if 'extensions' not in j2config or not j2config.get('extensions'):

          # make sure standard j2 extensions are added
          j2config['extensions'] = [
            'jinja2.ext.autoescape',
            'jinja2.ext.with_',
            'jinja2.ext.do',
            'jinja2.ext.loopcontrols'
          ] + (self.default_extensions or [])

        # auto-add hamlish extension
        for ext in self.syntax_extension:
          if ext not in j2config['extensions']:
            j2config['extensions'].append(ext)

        # factory environment
        environment = environment_factory(**j2config)

        if config.debug:
          environment.hamlish_mode, environment.hamlish_debug = 'indented', True

        # apply config overrides
        if 'TemplateAPI' in config.config and 'haml' in (
          config.config['TemplateAPI']):

          for (config_item, target_attr) in (
            ('mode', 'hamlish_mode'),
            ('extensions', 'hamlish_file_extensions'),
            ('div_shortcut', 'hamlish_enable_div_shortcut')):

            if config_item in config.config['TemplateAPI']['haml']:
              setattr(*(
                environment,
                target_attr,
                config.config['TemplateAPI']['haml'][config_item]))

        return environment

      syntax = _hamlish_syntax

    def environment(self, _handler, config):

      """ Prepare a new :py:class:`jinja2.Environment` object, for the purpose
          of rendering a template.

          :param _handler: Currently-active web handler. Always an instance of
            :py:class:`canteen.base.handler.Handler` or a subtype thereof.

          :param config: Reference to top-level application configuration
            object. Always an instance of :py:class:`canteen.util.Config`.

          :raises RuntimeError: If no template root path is configured.

          :returns: Instance of :py:class:`jinja2.Environment`, prepared with
            settings and configuration from the current runtime environment. """

      # grab template path, if any
      output = config.get('TemplateAPI', {'debug': True})
      jinja2_cfg = output.get('jinja2', self.default_config)

      _path = None
      if isinstance(config.app, dict):
        _path = config.app.get('paths', {}).get('templates', None)

      if not _path:
        # default path to cwd, and cwd + templates/, and cwd + templates/source
        cwd = os.getcwd()
        _path = (path.join(cwd),
                 path.join(cwd, 'templates'),
                 path.join(cwd, 'templates', 'source'))

      # shim-in our loader system, unless it is overriden in config
      if 'loader' not in jinja2_cfg:

        if output.get('force_compiled', False) or (
          isinstance(_path, dict) and 'compiled' in _path):

          if output.get('force_compiled', False):
            jinja2_cfg['loader'] = ModuleLoader(_path['compiled'], strict=True)
          else:
            choices = []
            if isinstance(_path, dict) and 'compiled' in _path:
              choices.append(ModuleLoader(_path['compiled']))

            if (isinstance(_path, dict) and 'source' in _path or (
                  isinstance(_path, basestring))):
              choices.append(FileLoader(_path['source'] if (
                             isinstance(_path, dict)) else _path))

            if not choices:  # pragma: no cover
              raise RuntimeError('No template path configured.')
            jinja2_cfg['loader'] = jinja2.ChoiceLoader(choices)

        else:
          jinja2_cfg['loader'] = FileLoader((
            _path['source'] if isinstance(_path, dict) else _path))

        if 'loader' not in jinja2_cfg:  # pragma: no cover
          raise RuntimeError('No configured template source path.')

      # make our new environment
      j2env = self.syntax(_handler, self.engine.Environment, jinja2_cfg, config)

      # allow jinja2 syntax overrides
      if 'syntax' in output:
        for override, directive in filter(lambda x: x[0] in output['syntax'], (
          ('block', ('block_start_string', 'block_end_string')),
          ('comment', ('comment_start_string', 'comment_end_string')),
          ('variable', ('variable_start_string', 'variable_end_string')))):

          # zip and properly set each group
          for group in zip(directive, output['syntax'][override]):
            setattr(j2env, *group)

      # add-in filters
      return j2env.filters.update(self.base_filters) or j2env

  @staticmethod
  def sanitize(content, _iter=True):

    """ Clean yielded template content of preceding and trailing whitespace,
        ensure correct encoding, and chunk/yield efficiently.

        :param content: Content to sanitize. Must be ``str`` or ``unicode``.

        :param _iter: ``bool`` flag indicating we'd like to return a generator
          rather than buffer and return the content directly.

        :returns: ``callable`` generator if ``_iter`` is truthy, otherwise a
          buffered ``list`` of ``content`` that has been sanitized. """

    # content should be a list of content blocks
    content = [content] if not (
      isinstance(content, (tuple, list))) else content

    def sanitize():

      """ Inner sanitization generator. Closured and provided by ``sanitize``
          so that content can be yielded as a generator. """

      # iteratively sanitize the response
      for _block in content: yield _block.strip()

    if _iter: return sanitize()  # return wrapped iterator
    return [block for block in sanitize()]

  @decorators.bind('template.base_context', wrap=property)
  def base_context(self):

    """ Prepare a set of default (*base*) template context items that should
        always be available (as *globals*) in the ``Jinja2`` render context.

        :returns: ``dict`` of ``(key => value)`` pairs to be injected as globals
          into ``Jinja2``'s rendering context. """

    from canteen.util import config

    return {
      # Python Builtins
      'all': all, 'any': any,
      'int': int, 'str': str,
      'len': len, 'map': map,
      'max': max, 'min': min,
      'enumerate': enumerate,
      'zip': zip, 'bool': bool,
      'list': list, 'dict': dict,
      'tuple': tuple, 'range': range,
      'round': round, 'slice': slice,
      'xrange': xrange, 'filter': filter,
      'reduce': reduce, 'sorted': sorted,
      'getattr': getattr, 'setattr': setattr,
      'unicode': unicode, 'reversed': reversed,
      '__debug__': __debug__ and config.Config().debug,
      'isinstance': isinstance, 'issubclass': issubclass}

  @decorators.bind('template.base_filters', wrap=property)
  def base_filters(self):

    """ Prepare a set of default (*base*) template filters that should be made
        available (as *globals*) in the ``Jinja2`` render context.

        :returns: ``dict`` of ``(key => value)`` pairs to be injected as
          globally-available ``Jinja2`` filters (usable as ``val|filter`` in
          templates, which results in ``filter(val)``). """

    return {

      # Python Builtins (besides the Jinja2 defaults, which are _awesome_)
      'len': len, 'max': max, 'maximum': max, 'min': min, 'minimum': min,
      'avg': average, 'average': average,  # support for basic averages
      'json': json.dumps, 'tojson': json.dumps  # support for json

    }  # @TODO(sgammon): markdown/RST support?

  @decorators.bind('template.render')
  def render(self, handler, config, template, ctx, _direct=False):

    """ Top-level render function that handles template render flow from start
        to finish. Loads template and performs render flow.

        :param handler: Currently-active web handler. Always an instance of
          :py:class:`canteen.base.handler.Handler` or a subtype thereof.

        :param config: Reference to global application configuration. Always an
          instance of :py:class:`canteen.util.config.Config`.

        :param template: ``str``/``unicode`` relative filepath to the template
          source file we'd like to render and return. Sometimes converted to a
          Python module path or DB lookup, depending on the active ``Loader``.

        :param ctx: Merged context to apply to ``Jinja2``'s rendering
          environment. Always a ``dict``, sometimes could be an empty one.
          Values made available in templates under the keys they are listed at.

        :param _direct: ``bool`` flag indicating we'd like the render flow to
          fully buffer and return, rather than passing a generator that can be
          consumed step-by-step.

        :returns: inner ``generator`` made by :py:meth`TemplateAPI.sanitize``,
         if ``_direct`` was falsy, otherwise a ``list`` of chunked template
         content entries. """

    # render template & return content iterato)
    content = (
      self.environment(handler, config).get_template(template).render(**ctx))

    # if _direct is requested, sanitize and roll-up buffer immediately
    if _direct: return self.sanitize(content, _iter=False)

    # otherwise, buffer/chain iterators to produce a streaming response
    return self.sanitize(content)



# -*- coding: utf-8 -*-

"""

  msgpack protocol tests
  ~~~~~~~~~~~~~~~~~~~~~~

  tests canteen's builtin msgpack RPC protocol.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
      A copy of this license is included as ``LICENSE.md`` in
      the root of the project.

"""

# canteen core, testing, model
from canteen.model import Model
from canteen.core import Library
from canteen.test import FrameworkTest


with Library('msgpack', strict=True) as (mlibrary, msgpack):

  import msgpack  # force re-import

  with Library('protorpc', strict=True) as (plibrary, protorpc):

    # load messages library
    messages = plibrary.load('messages')

    # msgapck protocol
    from canteen.rpc.protocol import msgpack as msgpackrpc


    class SampleMessage(messages.Message):

      """ Sample ProtoRPC message. """

      string = messages.StringField(1)
      integer = messages.IntegerField(2)


    class SampleParentMessage(messages.Message):

      """ Sample parent Message class. """

      sub = messages.MessageField(SampleMessage, 1)
      string = messages.StringField(2)
      integer = messages.IntegerField(3)


    class TestMsgpackRPCModel(Model):

      """ Sample ProtoRPC-integrated model """

      string = str
      integer = int


    class TestMsgpackRPCParentModel(Model):

      """ Sample ProtoRPC-integrated parent model. """

      sub = TestMsgpackRPCModel
      string = str
      integer = int


    class MsgpackProtocolTests(FrameworkTest):

      """ Tests `rpc.protocol.msgpack.Msgpack` """

      def test_msgpack_encode_message(self):

        """ Test encoding RPC messages in msgpack """

        # prepare message
        msg = SampleMessage(string='hi', integer=5)
        protocol = msgpackrpc.Msgpack()

        # encode message
        result = protocol.encode_message(msg)

        # decode and test
        inflated = msgpack.unpackb(result)

        # interrogate
        assert 'string' in inflated
        assert 'integer' in inflated
        assert inflated['string'] == 'hi'
        assert inflated['integer'] == 5

        return result, protocol

      def test_msgpack_decode_message(self):

        """ Test decoding RPC messages from msgpack """

        # prepare message
        encoded, protocol = self.test_msgpack_encode_message()

        # decode serialized body
        decoded = protocol.decode_message(SampleMessage, encoded)

        # interrogate
        assert decoded.string == 'hi'
        assert decoded.integer == 5

      def test_msgpack_encode_recursive(self):

        """ Test recursively encoding an RPC message using msgpack """

        s = SampleMessage(string='hiblab', integer=5)
        m = SampleParentMessage(string='hibleebs', integer=10, sub=s)

        assert s.integer == m.sub.integer == 5
        assert s.string == m.sub.string == 'hiblab'
        assert m.string == 'hibleebs'
        assert m.integer == 10
        assert m.sub is s
        assert isinstance(m, SampleParentMessage)
        assert isinstance(s, SampleMessage) and isinstance(m.sub, SampleMessage)

        # encode
        protocol = msgpackrpc.Msgpack()
        result = protocol.encode_message(m)

        assert result
        assert isinstance(msgpack.loads(result), dict)
        return result, s

      def test_msgpack_decode_recursive(self):

        """ Test recursively decoding an RPC message using msgpack """

        encoded, sub = self.test_msgpack_encode_recursive()
        protocol = msgpackrpc.Msgpack()

        # try decoding into full message
        result = protocol.decode_message(SampleParentMessage, encoded)

        assert sub.integer == result.sub.integer == 5
        assert sub.string == result.sub.string == 'hiblab'
        assert result.string == 'hibleebs'
        assert result.integer == 10
        assert result.sub
        assert isinstance(result, SampleParentMessage)
        assert isinstance(result.sub, SampleMessage)

      '''
      def test_msgpack_model_decode(self):

        """ Test decoding a msgpack RPC message into a `Model` """

        pass

      def test_msgpack_model_encode(self):

        """ Test encoding a `Model` into a msgpack RPC message """

        pass

      def test_msgpack_model_decode_recursive(self):

        """ Test recursively decoding an RPC with `Model` and msgpack """

        pass

      def test_msgpack_model_encode_recursive(self):

        """ Test recursively encoding an RPC with `Model` and msgpack """

        pass
      '''



# -*- coding: utf-8 -*-

"""

  graph model tests
  ~~~~~~~~~~~~~~~~~

  tests graph-extended models.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen model API
from canteen import model
from canteen.test import FrameworkTest

# default inmemory adapter
from canteen.model.adapter import inmemory


class TestPerson(model.Vertex):

  """ sample person (also a vertex) """

  firstname = basestring
  lastname = basestring


class TestTeammates(TestPerson > TestPerson):

  """ sample teammate-ship (edge that connects two people) """

  year_met = int


class VertexModelTests(FrameworkTest):

  """ Tests `model.Vertex`. """

  subject = inmemory.InMemoryAdapter()

  def test_construct(self):

    """ Test constructing a `Vertex` model """

    return TestPerson(
        key=model.VertexKey(TestPerson, 'sup'),
        firstname='John',
        lastname='Doe')

  def test_vertex_put(self):

    """ Test saving a `Vertex` model to storage """

    return self.test_construct().put(adapter=self.subject)

  def test_vertex_get(self):

    """ Test retrieving a `Vertex` by its key """

    assert TestPerson.get(self.test_vertex_put(), adapter=self.subject)

  def test_vertex_repr(self):

    """ Test string representation of a `Vertex` """

    p = TestPerson.get(self.test_vertex_put(), adapter=self.subject)
    assert 'Person' in repr(p)
    assert 'Person' in repr(TestPerson)
    assert 'Vertex' in repr(model.Vertex)


class EdgeModelTests(FrameworkTest):

  """ Tests `model.Edge`. """

  subject = inmemory.InMemoryAdapter()

  def test_spawn_directed(self):

    """ Test spawning directed `Edge` classes """

    class TestFriends(TestPerson > TestPerson):
        """ friend relationship """

    assert issubclass(TestFriends, model.Edge)
    assert not TestFriends.__spec__.directed

    class TestFriends(TestPerson < TestPerson):
        """ friend relationship """

    assert issubclass(TestFriends, model.Edge)
    assert not TestFriends.__spec__.directed

  def test_spawn_undirected(self):

    """ Test spawning undirected `Edge` classes """

    class Gift(TestPerson >> TestPerson):
        """ friend relationship """

    assert issubclass(Gift, model.Edge)
    assert Gift.__spec__.directed

    class Gift(TestPerson << TestPerson):
        """ friend relationship """

    assert issubclass(Gift, model.Edge)
    assert Gift.__spec__.directed

  def test_construct_undirected(self):

    """ Test constructing an undirected `Edge` model """

    # sam + alex
    sam, alex = TestPerson(firstname='Sam'), TestPerson(firstname='Alex')

    # low-level edge construct
    sam_to_alex = TestTeammates(sam, alex,
                                key=model.EdgeKey(TestTeammates, 'sup'),
                                year_met=2003)

    return sam_to_alex

  def test_construct_directed(self):

    """ Test constructing a directed `Edge` model """

    # ian + david
    ian, david = TestPerson(firstname='Sam'), TestPerson(firstname='Alex')

    # low-level edge construct
    ian_to_david = TestTeammates(ian, david,
                                 key=model.EdgeKey(TestTeammates, 'sup'),
                                 year_met=2003)

    return ian_to_david

  def test_undirected_edge_put(self):

    """ Test saving an undirected `Edge` model """

    return self.test_construct_undirected().put(adapter=self.subject)

  def test_directed_edge_put(self):

    """ Test saving a directed `Edge` model """

    return self.test_construct_directed().put(adapter=self.subject)

  def test_undirected_edge_get(self):

    """ Test retrieving an undirected `Edge` by its key """

    assert TestTeammates.get(self.test_undirected_edge_put(),
                             adapter=self.subject)

  def test_directed_edge_get(self):

    """ Test retrieving a directed `Edge` by its key """

    assert TestTeammates.get(self.test_directed_edge_put(),
                             adapter=self.subject)



# -*- coding: utf-8 -*-

"""

  init tests
  ~~~~~~~~~~

  tests things at the top-level package init for canteen.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen
import canteen
from canteen import test


class BaseFrameworkTests(test.FrameworkTest):

  """ Tests basic framework details. """

  def test_framework_all(self):

    """ Test for expected Framework-level exports """

    for attr in ('__all__',
                  'base', 'core', 'logic', 'model', 'rpc', 'runtime', 'util',
                  'Library', 'Logic', 'Page', 'Service', 'Model', 'Vertex',
                  'Edge', 'Key'):
      assert hasattr(canteen, attr), (
        "failed to resolve expected framework export: '%s'." % attr)



# -*- coding: utf-8 -*-

"""

  runscript
  ~~~~~~~~~

  accepts calls to ``canteen`` as a module. can be run with
  ``python -m canteen`` or ``python -m canteen/``, the latter
  assuming you have it installed right next to you.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# utils
from .util import walk
from .dispatch import (run, app)

walk(), run()



# -*- coding: utf-8 -*-

"""

  util
  ~~~~

  low-level utilities and miscellaneous tools that don't really
  belong anywhere special.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

from __future__ import print_function

# stdlib
import pkgutil

# submodules
from .cli import *
from .debug import *
from .config import *
from .struct import *
from .decorators import *


def say(*args):  # pragma: no cover

  """ Simply stringify and print the positional arguments received as ``args``.

      :param args: Positional arguments to ``repr`` and print.
      :returns: ``None``. """

  print(' '.join(map(lambda x: repr(x), args)))


def walk(root=None, debug=__debug__):  # pragma: no cover

  """ Force-walk all packages from ``root`` downward. DANGER: WILL IMPORT ALL
      MODULES BY FORCE, possibly loading code you didn't mean to load, or
      triggering scripts that aren't properly guarded against ``__main__``.

      :param root: Path to start walking packages from. Defaults to ``.``,
        meaning the current working directory.

      :param debug: ``bool`` debug flag - if ``True``, will print information
        about imported/found modules as it goes, and will raise exceptions
        instead of printing information about them.

      :returns: ``None``.  """

  # make sure working directory is in path
  if os.getcwd() not in sys.path:
    sys.path.insert(0, os.getcwd())

  def walker(bundle):  # pragma: no cover

    """ Inner function for recursively walking down a bundle. Print errors as
        they are encountered, raising them if ``debug`` is active.

        :param bundle: Current module bundle to work with, passed in from
          ``pkgutil.walk_packages`` in the form of ``loader``, ``name``, and
          ``is_package``.

        :returns: ``None``. """

    loader, name, is_package = bundle

    try:
      return importlib.import_module(name).__name__ if not is_package else name
    except ImportError as e:  # pragma: no cover
      print('Failed to preload path "%s"...' % (root or '.'))
      print(e)
      if debug: raise

  if debug: print('Preloading path "%s"...' % (root or '.'))
  say_lambda = (lambda x: say('Preloaded:', x) if x else None) if (
      debug) else (lambda x: x)
  map(say_lambda,
      map(walker, pkgutil.walk_packages(root or '.')))


__all__ = ('walk',
           'say',
           'cli',
           'config',
           'debug',
           'decorators',
           'struct',
           'bind')



# -*- coding: utf-8 -*-

"""

  RPC
  ~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import abc
import copy

# canteen
from canteen import core
from canteen import base
from canteen import model

# canteen core
from canteen.core import runtime
from canteen.core import injection

# canteen HTTP
from canteen.logic import http

# canteen util
from canteen.util import decorators
from canteen.util import struct as datastructures


with core.Library('protorpc', strict=True) as (library, protorpc):

  #### ==== Dependencies ==== ####

  # remote / message packages
  from protorpc import remote as premote
  from protorpc import registry as pregistry

  # message packages
  from protorpc import messages as pmessages
  from protorpc.messages import Field as ProtoField
  from protorpc.messages import Message as ProtoMessage

  # message types
  from protorpc import message_types as pmessage_types
  from protorpc.message_types import VoidMessage as ProtoVoidMessage

  # WSGI internals
  from protorpc.wsgi import util as pwsgi_util
  from protorpc.wsgi import service as pservice


  ## Globals
  _RPC_BASE_URI = '/_rpc/'


  #### ==== Message Fields ==== ####

  class VariantField(ProtoField):

      """ Field definition for a completely variant field. Allows containment
          of any valid Python value supported by Protobuf/ProtoRPC. """

      VARIANTS = frozenset([pmessages.Variant.DOUBLE, pmessages.Variant.FLOAT,
                            pmessages.Variant.BOOL, pmessages.Variant.INT64,
                            pmessages.Variant.UINT64, pmessages.Variant.SINT64,
                            pmessages.Variant.INT32, pmessages.Variant.UINT32,
                            pmessages.Variant.SINT32, pmessages.Variant.STRING,
                            pmessages.Variant.BYTES, pmessages.Variant.MESSAGE,
                            pmessages.Variant.ENUM])

      DEFAULT_VARIANT = pmessages.Variant.STRING

      type = (int, long, bool, basestring, dict, pmessages.Message)


  class StringOrIntegerField(ProtoField):

      """ Field definition for a field that can contain either a string or
          integer. Usually used for key names/IDs or message IDs/hashes. """

      VARIANTS = frozenset([pmessages.Variant.STRING, pmessages.Variant.DOUBLE,
                            pmessages.Variant.INT64, pmessages.Variant.INT32,
                            pmessages.Variant.UINT64, pmessages.Variant.UINT32])

      DEFAULT_VARIANT = pmessages.Variant.STRING

      type = (int, long, basestring, dict, pmessages.Message)


  #### ==== Message Classes ==== ####

  class Key(ProtoMessage):

      """ Message for a :py:class:`canteen.model.Key`. """

      encoded = pmessages.StringField(1)  # encoded (`urlsafe`) key
      kind = pmessages.StringField(2)  # kind name for key
      id = StringOrIntegerField(3)  # integer or string ID for key
      namespace = pmessages.StringField(4)  # string namespace for key
      parent = pmessages.MessageField('Key', 5)  # recursive key message


  class Echo(ProtoMessage):

      """ I am rubber and you are glue... """

      message = pmessages.StringField(1, default='Hello, world!')


  ## expose message classes alias
  messages = datastructures.WritableObjectProxy(**{

      # canteen-provided messages
      'Key': Key,  # message class for a canteen model key
      'Echo': Echo,  # echo message defaulting to `hello, world` for testing

      # builtin messages
      'Message': ProtoMessage,  # top-level protorpc message class
      'VoidMessage': ProtoVoidMessage,  # top-level protorpc void message

      # specific types
      'Enum': pmessages.Enum,  # enum descriptor / definition class
      'Field': pmessages.Field,  # top-level protorpc field class
      'FieldList': pmessages.FieldList,  # top-level protorpc field list class

      # field types
      'VariantField': VariantField,  # generic hold-anything property
      'BooleanField': pmessages.BooleanField,  # boolean true/false field
      'BytesField': pmessages.BytesField,  # low-level binary-safe string field
      'EnumField': pmessages.EnumField,  # field for referencing an `Enum` class
      'FloatField': pmessages.FloatField,  # field for a floating point number
      'IntegerField': pmessages.IntegerField,  # field for an integer
      'MessageField': pmessages.MessageField,  # field for a sub-message
      'StringField': pmessages.StringField,  # field for unicode or ASCII
      'DateTimeField': pmessage_types.DateTimeField  # field for datetime types

  })


  def service_mappings(services, registry_path='/_rpc/meta', protocols=None):

    """ Generates mappings from `url -> service` for registered Canteen RPC
        services.

        Takes an iterable of URL and service mappings, wraps with appropriate
        WSGI utilities, and registers with registry service for Endpoints/meta
        integration.

        :param services: Iterable of services, preferably a ``list`` of
          ``tuples``, where each is in the format ``(url, service)``. ``url``
          should be a relative prefix for matching requests, like
          ``/_rpc/hello`` for something called ``HelloService``.

        :param registry_path: Path prefix for ``RegistryService``, which returns
          metadata about registered RPC services. Required for integration with
          Google Cloud Endpoints or the various ProtoRPC client-side library
          generation options out there.

        :param protocols: Protocols to use for dispatching services. Custom
          protocol implementations are supported and two are shipped with
          canteen - ``JSON`` and ``msgpack`` RPC formats (note: not necessarily
          affiliated with any standards that are actually called "msgpack-rpc"
          or "jsonrpc").

        :returns: WSGI application prepared by :py:mod:`protorpc`, which, upon
          dispatch, will attempt to delegate response to the first matching
          ``Service`` implementation, as governed by the mappings generated in
          this function from ``services``. """

    if not protocols:
      # load canteen builtin protocols
      from canteen.base import protocol
      protocols = protocol.Protocol.mapping

    if isinstance(services, dict):
      services = services.iteritems()

    final_mapping, paths, registry_map = (
      [],
      set(),
      {} if registry_path else None
    )

    for service_path, service_factory in services:
      service_class = service_factory.service_class if (
        hasattr(service_factory, 'service_class')) else service_factory

      if service_path not in paths:
        paths.add(service_path)
      else:
        raise premote.ServiceConfigurationError(
          'Path %r is already defined in service mapping' %
          service_path.encode('utf-8'))

      if registry_map is not None: registry_map[service_path] = service_class
      final_mapping.append((
        pservice.service_mapping(*(
          service_factory, service_path), protocols=protocols)))

    if registry_map is not None:
      final_mapping.append(pservice.service_mapping(*(
        pregistry.RegistryService.new_factory(registry_map), registry_path),
            protocols=protocols))

    return pwsgi_util.first_found(final_mapping)


  @http.url('rpc', (
    r'%s<string:version>/<string:service>.<string:method>' % _RPC_BASE_URI))
  class ServiceHandler(base.Handler):

    """ Builtin concrete :py:class:`base.Handler` for use with RPC services. As
        services are bound to names, they are registered here and eventually
        mapped URLs are generated (via `service_mappings`).

        Normally this handler is mapped at ``<version>/<service>.<method>``,
        which supports both service types ('concrete' and 'registry') at the
        following URLs (with examples inline):

        - concrete: ``v1/hello.hi`` for a ``HelloService`` with ``hi`` method
        - meta: ``meta/registry.services`` to describe a service's methods """

    __services__ = {}  # holds services mapped to their names

    @classmethod
    def add_service(cls, name, service, config=None, **kwargs):

      """ Add a service to this handler's local dispatch registry.
          Called from ``@rpc.service`` to mount a service to dispatch.

          :param name: Simple string name for the service. For instance,
            ``hello`` for ``HelloService``.

          :param service: Service class to be registered.

          :param config: Configuration can be passed as a dictionary
            (at ``config``) or with ``kwargs``, which override items in
            ``config``.

          :returns: The service class passwd at ``service``. """

      config = config or {}
      config.update(kwargs)
      cls.__services__[name] = (service, config or {})
      return service

    @decorators.classproperty
    def services(cls):

      """ Iterator for all locally-registered services, presented as
          a class-level property.

          :yields: Each named service, in the tupled format
            ``name, service``, much like ``dict.iteritems``. """

      for name in sorted(cls.__services__.iterkeys()):
        yield name, cls.__services__[name]

    @classmethod
    def get_service(cls, name):

      """ Retrieve a locally-registered service by name.

          :param name: Short name for the service. For instance,
            ``hello`` for ``HelloService``.

          :returns: Registered ``rpc.Service`` class at that name,
            or ``None`` if no matching service could be located. """

      if name in cls.__services__:
        return cls.__services__[name][0]

    @classmethod
    def describe(cls, json=False,
                      javascript=False,
                      callable='apptools.rpc.service.factory'):

      """ Describe locally-registered services in various formats.
          Exposed to template context as ``services.describe``, so that
          frontends can easily be notified of supported RPC services.

          Omitting both ``json`` and ``javascript`` will return a ``list``
          of ``tuples`` describing each service, as ``(name, methods, config)``.

          Passing ``json`` *and* ``javascript`` is unsupported and raises
          a ``TypeError`` describing your foolishness.

          :param json: Describe the services as a JSON string, suitable for
            placement on an HTML page. Boolean, defaults to ``False``.

          :param javascript: Generate JS that calls a function (assumed to
            already be present on ``window``) with a structure describing
            locally registered services, suitable for placement on an HTML
            page. Boolean, defaults to ``False``.

          :param callable: Opportunity to change the frontend callable function
            that will be passed the service manifest. Defaults to the hard-coded
            value ``apptools.rpc.service.factory`` for backwards compatibility.

          :returns: ``list`` of ``tuples`` if requesting structured description,
            or a JSON string of that structure if ``json=True`` is passed, or JS
            code invoked with that JSON structure if ``javascript=True`` is
            passed. """

      _services = []
      for name, service in cls.services:
        service, config = service
        _services.append((
          name,  # service shortname

          # service methods
          tuple((name for name in service.all_remote_methods().iterkeys())),

          # service config
          config or {}
        ))

      if json and javascript:
        raise TypeError('Please pick between "JSON" and "JavaScript"'
                        ' output for services.')

      if json:  # generate JSON only?
        import json as serializer
        return serializer.dumps(_services)

      if javascript:  # generate javascript?
        import json as serializer
        return "%s(%s);" % (callable, serializer.dumps(_services))
      return _services  # or return raw?

    @decorators.classproperty
    def application(cls):

      """ Utility for generating a WSGI application capable of dispatching
          locally-registered services on ``ServiceHandler``, exposed as a class-
          level property.

          Uses :py:mod:`protorpc`'s fantastic `wsgi.utils.first_found`, which
          will dispatch `rpc.Service` applications one at a time until a non-404
          error occurs, in which case response is delegated to that application.

          If no application can be found to match the given WSGI state, an
          ``HTTP 404`` is raised.

          :returns: Prepared ``protorpc.wsgi.utils.first_found`` WSGI
            application closure.  """

      _services = []
      for name, service in cls.services:

        service, config = service
        service_factory = service.new_factory(config=config)

        # Update docstring so that it is easier to debug.
        full_class_name = '%s.%s' % (service.__module__, service.__name__)
        if __debug__:
          service_factory.func_doc = (
              'Creates new instances of service %s.\n\n'
              'Returns:\n'
              '  New instance of %s.'
              % (service.__name__, full_class_name))

        # Update name so that it is easier to debug the factory function.
        service_factory.func_name = '%s_service_factory' % service.__name__

        service_factory.service_class = service

        # resolve service version
        _version = config.get('version', 'v1')
        if isinstance(_version, int): _version = 'v' + str(_version)
        _services.append((r'/_rpc/%s/%s' % (_version, name), service_factory))

      return service_mappings(_services, registry_path='/_rpc/meta/registry')

    def OPTIONS(self, version, service, method):

      """ Dispatch handler for ``HTTP OPTIONS`` requests. Specifies available
          HTTP methods as ``OPTIONS`` and ``POST``, since (by default) only
          RPC-like functionality is supported.

          :param version: Service version, as specified in the URL.
          :param service: Service name, as specified in the URL.
          :param method: Service method to dispatch, as specified in the URL.

          :returns: Suitable HTTP response for an ``HTTP OPTIONS`` request. """

      # @TODO(sgammon): HTTP OPTIONS integration testing for services
      return self.response('OPTIONS, POST')  # pragma: no cover

    def POST(self, version, service, method):  # pragma: no cover

      """ Dispatch handler for ``HTTP POST`` requests. Main entrypoint into
          the RPC framework via HTTP, as requests are ``POST``ed in a supported
          ``Content-Type`` for an attached ``Protocol``.

          :param version: Service version, as specified in the URL.
          :param service: Service name, as specified in the URL.
          :param method: Service method to dispatch, as specified in the URL.

          :returns: Suitable HTTP response for an ``HTTP POST`` request.  """

      # @TODO(sgammon): HTTP POST integration testing for services
      _status, _headers = None, None

      def _respond(status, headers):  # pragma: no cover

        """ Inner response closure that overrides normal ``start_response``.

            :param status: HTTP status for response.
            :param headers: Iterable of ``(name, value)`` pairs for HTTP
            response headers. """

        _status, _headers = status, headers

      # delegate to service application
      return self.response.__class__((
        self.application(self.environment, _respond)), **{
        'status': _status,
        'headers': _headers
      })

    GET = POST


  class Exception(premote.ApplicationError):

      """ Base RPC exception class, used to indicate an application-level error.
          Backed by ProtoRPC's :py:class:`remote.ApplicationError`. """


  class ServerException(premote.ServerError):

      """ Base RPC exception class, used to indicate a server-side error.
          Backed by ProtoRPC's :py:class:`remote.ServerError`. """


  class ClientException(premote.RequestError):

      """ Base RPC exception class, used to indicate a client-side error. Backed
          by ProtoRPC's :py:class:`remote.RequestError`. """


  class Exceptions(datastructures.ObjectProxy):

      """ Datastructure used to house remote-capable RPC exceptions, and for
          expression of error types in meta APIs. """


  class AbstractService(premote.Service):

    """ Abstract class that provides basic Canteen integration for remote
        services. Handles MRO injection and ABC enforcement."""

    class __metaclass__(premote.Service.__metaclass__):

      """ Inline metaclass for `AbstractService` that injects a DI delegate into
          the MRO chain. """

      __delegate__ = None  # dependency injection delegate class

      def mro(cls):

        """ Inject a DI delegate (generated at `cls.delegate`) to enable
            MRO-based injection.

            :returns: Assembled MRO chain (a ``list``). """

        # noinspection PyCallByClass
        chain = type.mro(cls)

        if not premote.StubBase in cls.__bases__:
          if cls.__name__ is "AbstractService":
            # wrap delegate deep in the root
            return chain[0:-1] + [cls.delegate()] + chain[-1:]
        return chain  # it's a stub or something else - don't touch anything

      def delegate(cls):

        """ Generate a DI delegate suitable for use as a window into the main
            DI pool from ``AbstractService``.

            :returns: Bound :py:class:`injection.Delegate`. """

        cls.__class__.__delegate__ = injection.Delegate.bind(cls)
        return cls.__class__.__delegate__

    @abc.abstractproperty
    def exceptions(self):  # pragma: no cover

      """ Abstract property requiring the specification of expected exceptions
          from a remote service.

          :raises NotImplementedError: Always, as this property is abstract. """

      raise NotImplementedError('Property `AbstractService.exceptions`'
                                ' requires implementation by a concrete'
                                ' subclass and cannot be invoked directly.')


  class Service(AbstractService):

    """ Base remote service class, used as an extension point for Canteen
        application developers to create their own remotely-accessible RPC
        services. """

    __state__ = None  # local state
    __config__ = None  # local configuration

    def __init__(self, config=None, **kwargs):

      """ Initialize a new ``Service`` instance, optionally overlaying
          configuration with ``config`` or ``kwargs``, which override items in
          ``config``.

          :param config: Positional opportunity to specify service
            configuration. """

      config = config or {}
      config.update(kwargs)
      self.__config__ = config

    ## = property accessors = ##
    state = property(lambda self: self.__state__)  # accessor for __state__
    config = property(lambda self: self.__config__)  # accessor for __config__
    platform = property(lambda self: self.__bridge__)  # accessor for __bridge__

    @classmethod
    def new_factory(cls, *args, **kwargs):

      """ Classmethod for constructing a new ``ServiceFactory`` to wrap this
          ``Service`` class. Passes positional and keyword arguments for later
          use in ``Service.__init__``.

          :returns: Constructed ``ServiceFactory`` instance. """

      return ServiceFactory.construct(cls, *args, **kwargs)

    def initialize_request_state(self, state):  # pragma: no cover

      """ Initialize this service with ``state`` handed-in by ProtoRPC's
          underlying plumbing.

          :param state: State to attach locally. Always a descendent of
            :py:class:`remote.RequestState`, and usually an instance of
            :py:class:`remote.HttpRequestState` (which is used when running over
            HTTP, obviously).

          :returns: Nothing. """

      # @TODO(sgammon): full-cycle HTTP RPC testing
      self.__state__ = state
      if hasattr(self, 'initialize'):
        try:  # @TODO(sgammon): better logging here
          self.initialize(state)  # hand down to initialize hook
        except Exception as e:
          import traceback; traceback.print_exc()
          if __debug__: raise


  class ServiceFactory(object):

    """ Specifies an object that wraps an ``rpc.Service`` class and encapsulates
        a static set of ``args`` and ``kwargs`` to construct instaces of that
        wrapped ``Service``. """

    service = Service  # service class to factory
    args, kwargs = None, None  # service init args

    def __new__(cls, **kwargs):

      """ New instances requested of this ``cls`` are redirected to the
          constructor for the locally encapsulated ``Service`` class.

          Updated ``kwargs`` can be passed via this constructor, to be overlayed
          on ``cls``-level ``kwargs`` before being handed to the target's
          constructor.

          :returns: Constructed ``Service`` object. """

      _kargs = copy.deepcopy(cls.kwargs)
      _kargs.update(kwargs)

      return cls.service(*(cls.args or tuple()), **(_kargs or {}))

    @classmethod
    def construct(cls, service, *args, **kwargs):

      """ Construct a new, dynamic ``ServiceFactory`` subclass dedicated to
          wrapping ``service``, optionally specifying positional ``args`` and
          keyword ``kwargs``.

          :param service: ``Service`` class to wrap with the newly-constructed
            ``ServiceFactory`` subclass.

          :returns: Dynamically-generated ``ServiceFactory`` class wrapping
            given ``service`` class. """

      return type(service.__name__ + 'Factory', (cls,), {
        'args': args,
        'kwargs': kwargs,
        'service': service
      })

    # shortcut to the ``service_class``
    service_class = decorators.classproperty(lambda cls: cls.service)

    # shortcut to alias f() to s()
    __call__ = lambda self: self.service(*self.args, **self.kwargs)


  class remote(object):

    """ Utility class that provides utilities for registering and mounting
        ``remote`` methods, services, and configuration. """

    name = None  # string name for target
    config = None  # config items for target
    target = None  # contains a service if wrapping one

    def __init__(self, name, expose='public', **config):

      """ Initialize this instance of ``remote`` with a string ``name`` and
          exposure policy (at ``expose``, which defaults to 'public').

          Configuration can be passed as kwargs. """

      self.name, self.config = name, config

    @classmethod
    def register(cls, name_or_message, response=None, **config):

      """ Internal function used to register this remote object, bound to either
          a ``name`` (if the target is a ``Service``) or a ``request`` and
          ``response`` message pair.

          Configuration can be passed as kwargs.

          :returns: Closured function designed to wrap the target callable with
            a routine to register it before construction. """

      if isinstance(name_or_message, basestring):
        name, request = name_or_message, None
      else:
        name, request = None, name_or_message

      if not name:

        request_klass, response_klass = None, None
        if isinstance(request, type) and issubclass(request, model.Model):
          request_klass = response_klass = request.to_message_model()

        if response and response != request:
          if isinstance(response, type) and issubclass(response, model.Model):
            response_klass = response.to_message_model()

        request_klass, response_klass = (
          request_klass or name_or_message,
          response_klass or (response or name_or_message)
        )

        def _remote_method(method):

          """ Closure to wrap the target ``method`` at construction time with
              appropriate tooling to execute remote RPCs.

              :param method: Remote-capable method to be wrapped.

              :returns: Wrapped closure ``_respond`` that, when called,
                dispatches the target ``method``. """

          def _respond(self, _request_message):

            """ Inner closure designed to wrap the raw remote ``method`` and
                enforce validation/conversion of ProtoRPC types to native Python
                ones.

                Also handles conversion in/out of Canteen models, if the remote
                method so chooses, by binding to ``canteen.Model`` subclasses
                instead of ProtoRPC ``message.Message``s.

                :param _request_message: Remotely-submitted ``request`` to
                (potentially) be converted and submitted to inner service
                ``method``.

                :returns: Result of calling ``method`` with ``_request_message``
                  as the originating request, so long as the request was
                  successfully executed and passed all client-related
                  constraints. """

            if isinstance(request, type) and issubclass(request, model.Model):
              # convert incoming message to model
              result = method(self, request.from_message(_request_message))

            else:
              # we're using regular messages always
              result = method(self, _request_message)

            # convert outgoing message to model if it isn't already
            if isinstance(result, model.Model):
              return result.to_message()
            return result

          # wrap responder
          wrapped = premote.method(request_klass, response_klass)(_respond)

          # make things transparent
          wrapped.__name__, wrapped.__doc__, wrapped.__inner__ = (
            method.__name__,
            method.__doc__,
            method)

          _respond.__inner__ = wrapped

          # quack quack
          _respond.__name__, _respond.__doc__, _respond.remote = (
            method.__name__,
            method.__doc__,
            wrapped.remote)

          # just for backup
          wrapped.__remote_name__, wrapped.__remote_doc__ = (
            method.__name__,
            method.__doc__,)

          # add remote info
          wrapped.__remote__ = wrapped.remote

          return _respond
        return _remote_method

      # wrap wrap wrapper
      config['expose'] = config.get('expose', 'public')
      return cls(name, **config)

    # shorthand for `cls.register` with `expose=public`
    public = lambda cls, *args, **config: (
      cls.register(*args, expose='public', **config))

    # shorthand for `cls.register` with `expose=private`
    private = lambda cls, *args, **config: (
      cls.register(*args, expose='private', **config))

    # aliases for `cls.register`
    method = service = register

    def __call__(self, target):

      """ Callable wrapped for ``Service`` classes that is dispatched as part of
          the construction/decorator flow. Adds hooks to local
          ``ServiceHandler`` for the locally-bound ``target`` ``Service`` class.

          :param target: ``Service`` class to be registered and bound.

          :returns: ``target``, after executing hooks and registering. """

      self.target = target

      # finally, register the service (if it's a service class)
      if isinstance(target, type) and issubclass(target, Service):

        # call service registration hooks
        runtime.Runtime.execute_hooks('rpc-service', service=target)

        # call method registration hooks
        for method in target.all_remote_methods():  # pragma: no cover
          runtime.Runtime.execute_hooks('rpc-method',
                                        service=target,
                                        method=method)

        ServiceHandler.add_service(self.name, target, **self.config)

      return target


  __all__ = ('Service',
             'remote',
             'ServiceHandler',
             'service_mappings',
             'messages',
             'Key',
             'Echo',
             'VariantField',
             'protocol',
             'exceptions')



# -*- coding: utf-8 -*-

"""

  msgpack RPC protocol
  ~~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen base & core
from canteen.core import runtime
from canteen.base import protocol


## Globals
_content_types = (
  'application/msgpack',
  'application/x-msgpack')


with runtime.Library('msgpack') as (msglib, msgpack):
  with runtime.Library('protorpc') as (protolib, protorpc):

    # submodules
    protojson = protolib.load('protojson')
    messages = protolib.load('messages')


    @protocol.Protocol.register('msgpack', _content_types)
    class Msgpack(protocol.Protocol, protojson.ProtoJson):

      """  """

      def encode_message(self, message):

        """  """

        message.check_initialized()

        def _walk_struct(m):

          """ recursively encode msgpack """

          # simple extraction
          _packed = {}
          for field in m.all_fields():
            value = getattr(m, field.name)
            if value is not None:
              if isinstance(value, messages.Message):
                _packed[field.name] = _walk_struct(value)
              else:
                _packed[field.name] = value
          return _packed
        return msgpack.packb(_walk_struct(message))

      def decode_message(self, message_type, encoded_message):

        """  """

        # garbage in, garbage out
        if not encoded_message.strip():  # pragma: no cover
          return message_type()

        dictionary = msgpack.unpackb(encoded_message)
        message = self._ProtoJson__decode_dictionary(message_type, dictionary)
        message.check_initialized()
        return message



# -*- coding: utf-8 -*-

"""

  protocol base
  ~~~~~~~~~~~~~

  ``Protocol`` classes help the framework understand different serialization
  dialects for use in RPC. For example, the JSON format used for most browser
  is specified as a ``Protocol``.

  As is customary with :py:mod:`protorpc` (where the equivalent class used to
  be called a ``Mapper`` and is now just a sloppy collection of ``object``s),
  ``Protocol``s register ``Content-Type``s to respond to. When an RPC message
  matching that content type is submitted, the proper ``Protocol`` is used to
  decode it.

  On the client side, ``Protocol`` objects can be used by handing in a custom
  object when creating a ``Transport``.

  Example:

    # -*- coding: utf-8 -*-
    import json
    from canteen import Protocol

    @Protocol.register('jsonrpc', ('application/json', 'text/json'))
    class JSONRPC(Protocol):

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import abc

# canteen util & core
from ..core import runtime
from ..util import decorators


with runtime.Library('protorpc') as (library, protorpc):

  # protorpc
  messages, protojson, remote = library.load(*(
    'messages',
    'protojson',
    'remote'))


  # noinspection PyMethodParameters
  class Protocol(object):

    """ Base ``Protocol`` class for adding serialization dialects to the RPC
        engine. Implementing subclasses is easy - just specify an encoder and
        decoder function at ``encode_message`` and ``decode_message``.

        Once a ``Protocol`` class is written, it can be registered with the RPC
        subsystem by decorating it with ``Protocol.register``, along with a
        short string name (for instance, ``jsonrpc``) and a set of content types
        to respond to. """

    __label__ = None  # string (human) name for the protocol
    __config__ = None  # local configuration info
    __protocols__ = {}  # class-level map of all protocols to their names
    __metaclass__ = abc.ABCMeta  # enforce ABC compliance
    __content_types__ = ('',)  # default to no matched content types

    ## == Multi-protocol Tools == ##

    @decorators.classproperty
    def all(cls):

      """ Class-level generator accessor to iterate through all registered
          ``Protocol`` implementations. Used at construction time to find
          protocol classes to bundle into a :py:mod:`protorpc` ``Protocols``
          object.

          :param cls: Current ``cls``, as this is a class-level property.

          :returns: Yields ``Protocol`` implementations one at a time. """

      for protocol in cls.__protocols__.itervalues():
        yield protocol

    @decorators.classproperty
    def mapping(cls):

      """ Class-level accessor for creating a :py:mod:`protorpc` ``Protocols``
          container, which resolves the proper ``Protocol`` object to dispatch
          an RPC with, based on the HTTP ``Content-Type`` header.

          :param cls: Current ``cls``, as this is a class-level property.

          :returns: :py:class:`protorpc.remote.Protocols` object containing all
            registered :py:class:`Protocol` objects. """

      container = remote.Protocols()  # construct a protocol container

      for protocol in cls.all:

        # construct protocol singleton and add to container
        singleton = protocol()

        container.add_protocol(*(
          singleton,
          singleton.name,
          singleton.content_type,
          singleton.alternative_content_types))

      return container

    ## == Registration Decorator == ##
    @classmethod
    def register(cls, name, types, **config):

      """ Register a ``Protocol`` implementation by name and a set of content
          types. Usually used as a decorator.

          ``kwargs`` are taken as configuration to add to the protocol target.

          :param name: Name to register this protocol under. ``str`` or
            ``unicode``.

          :param types: Content types to use this mapper for. Iterable of
            ``str`` or ``unicode`` ``Content-Type`` values to match on.

          :param config: Keyword arguments are accepted as extra config to be
            passed to the underlying ``Protocol`` subclass.

          :returns: Closured wrapper to register the protocol and return it. """

      assert isinstance(name, basestring), "protocol name must be a string"
      assert isinstance(types, (list, tuple)), "types must be an iterable"

      def _register_protocol(klass):

        """ Closure to register a class as an available protocol right before
            class construction. Mounts a ``Protocol``'s label, content types,
            and configuration.

            :param klass: Target class to register and return. Subclass of
              :py:class:`canteen.base.protocol.Protocol`.

            :returns: Target ``klass``, after registration. """

        # assign protocol details
        klass.__label__, klass.__content_types__, klass.__config__ = (
          name,
          types,
          config)

        if name not in Protocol.__protocols__:
          Protocol.__protocols__[name] = klass  # register :)
        return klass
      return _register_protocol

    ## == Protocol Properties == ##

    @decorators.classproperty
    def name(cls):

      """ Class-level accessor for the 'short name' of this ``Protocol`` class.
          For example, ``jsonrpc``.

          :param cls: Current ``cls``, as this is a class-level property.

          :returns: ``str`` short name for this :py:class:`Protocol`. """

      return cls.__label__

    @decorators.classproperty
    def content_type(cls):

      """ Class-level accessor for the primary ``Content-Type`` this
          :py:class:`Protocol` should respond to. The first entry in the
          available options is used as the primary ``Content-Type``.

          :param cls: Current ``cls``, as this is a class-level property.

          :returns: Primary ``str`` ``Content-Type`` value. """

      return cls.__content_types__[0]

    @decorators.classproperty
    def alternative_content_types(cls):

      """ Class-level accessor for 'alternative' ``Content-Type``s that this
          :py:class:`Protocol` should respond to. Alternative content types
          will be responded to, but not used for responses (the 'primary'
          ``Content-Type`` is used as the response's type).

          :param cls: Current ``cls``, as this is a class-level property.

          :returns: ``list`` of ``str`` ``Content-Type``s. """

      return [i for i in (
        filter(lambda x: x != cls.content_type, cls.__content_types__))]

    ## == Abstract Methods == ##
    @abc.abstractmethod
    def encode_message(self, message):

      """ Encode a message according to this :py:class:`Protocol`. Must be
          implemented by child classes, and so is marked as an abstract method.

          Failure to specify this method will prevent an implementing class
          from being constructed.

          :param message: Object to encode with ``Protocol``. Passed-in from
            ProtoRPC as ``message.Message``.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: ``None``, but child class implementations are expected to
            return an encoded ``str``/``unicode`` representation of
            ``message``. """

      raise NotImplementedError('Method `Protocol.encode_message`'
                                ' is abstract.')  # pragma: no cover

    @abc.abstractmethod
    def decode_message(self, message_type, encoded_message):

      """ Decode a message according to this :py:class:`Protocol`. Must be
          implemented by child classes, and so is marked as an abstract method.

          Failure to specify this method will prevent an implementing class from
          being constructed.

          :param message_type: Type to expand from ``encoded_message``. Subclass
            of :py:class:`messages.Message`.

          :param encoded_message: Encoded message to be decoded. ``str``,
            ``unicode`` or iterable string buffer.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: ``None``, but child class implementations are expected to
            return an inflated ``message_type`` based on the provided
            ``encoded_message``. """

      raise NotImplementedError('Method `Protocol.decode_message`'
                                ' is abstract.')  # pragma: no cover

    CONTENT_TYPE = content_type
    ALTERNATIVE_CONTENT_TYPES = alternative_content_types



# -*- coding: utf-8 -*-

"""

  protorpc adapter tests
  ~~~~~~~~~~~~~~~~~~~~~~

  tests integration between canteen's model layer and
  Google's :py:mod:`protorpc`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import datetime

# testing
from canteen import rpc
from canteen import test
from canteen import model
from canteen import struct

# protorpc
from protorpc import messages

# adapter & exceptions
from canteen.model import exceptions
from canteen.model.adapter import protorpc


class ProtoRPCAdapterModuleTest(test.FrameworkTest):

  """ Tests the ProtoRPC model adapter """

  def test_module_globals(self):

    """ Test ProtoRPC adapter globals """

    assert hasattr(protorpc, 'protorpc')
    assert hasattr(protorpc, 'pmessages')
    assert hasattr(protorpc, 'pmessage_types')
    assert hasattr(protorpc, 'build_message')
    assert hasattr(protorpc, 'ProtoRPCKey')
    assert hasattr(protorpc, 'ProtoRPCModel')

  def test_build_message(self):

    """ Test `build_message` with a basic `Model` """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=5)
    message = message_class(string='hi', integer=5)

    assert _model.string == message.string
    assert _model.integer == message.integer

  def test_build_message_default(self):

    """ Test `build_message` with a property that has a default value """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int, {'default': 10}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi')
    message = message_class(string='hi')

    assert _model.string == message.string
    assert _model.integer == message.integer

  def test_build_message_required(self):

    """ Test `build_message` with a required property """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int, {'required': True}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi')
    message = message_class(string='hi')

    assert _model.string == message.string
    assert _model.integer == message.integer

    with self.assertRaises(exceptions.PropertyRequired):
      _model.put()

    with self.assertRaises(messages.ValidationError):
      message.check_initialized()

  def test_build_message_repeated(self):

    """ Test `build_message` with a repeated property """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int, {'repeated': True}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert message_class.integer.repeated is True
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=[1, 2, 3])
    message = message_class(string='hi', integer=[1, 2, 3])

    assert _model.string == message.string
    assert type(_model.integer) is list
    assert type(message.integer) is messages.FieldList
    assert len(_model.integer) == 3 and len(message.integer) == 3

  def test_build_message_explicit_implementation_field(self):

    """ Test `build_message` with a valid explicit implementation field """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = float, {'field': 'IntegerField'}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=5.5)
    message = message_class(string='hi', integer=5)

    assert _model.string == message.string
    assert isinstance(_model.integer, float)
    assert isinstance(message.integer, int)

  def test_build_message_invalid_implementation_field(self):

    """ Test `build_message` with an invalid explicit implementation field """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int, {'field': 'NoSuchField'}

    with self.assertRaises(ValueError):
      protorpc.build_message(SimpleModel)

  def test_build_message_skip_field(self):

    """ Test `build_message` with an indication to skip a field """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = float, {'field': False}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert not hasattr(message_class, 'integer')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=5.5)
    message = message_class(string='hi')

    assert _model.string == message.string
    with self.assertRaises(AttributeError):
      message.integer

  def test_build_message_explicit_field_args(self):

    """ Test `build_message` with implementation field args """


    class SimpleEnum(messages.Enum):

      """ Enumerates colors! """

      RED = 0x0
      BLUE = 0x1
      GREEN = 0x2


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      color = basestring, {'field': ('EnumField', (SimpleEnum,))}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'color')
    assert message_class.color.type is SimpleEnum

  def test_build_message_with_enum(self):

    """ Test `build_message` with a message that contains an enum """


    class Color(struct.BidirectionalEnum):

      """ Sample enumeration of a bunch of colors. """

      BLUE = 0x0
      RED = 0x1
      GREEN = 0x2


    class SomeModel(model.Model):

      """ Something involving colors. """

      color = Color
      name = str

    message_class = protorpc.build_message(SomeModel)

    assert hasattr(message_class, 'color')
    assert hasattr(message_class, 'name')
    assert message_class.color.type.BLUE is Color.BLUE

  def test_build_message_explicit_field_args_kwargs(self):

    """ Test `build_message` with an implementation of field args + kwargs """


    class SimpleEnum(messages.Enum):

      """ Enumerates colors! """

      RED = 0x0
      BLUE = 0x1
      GREEN = 0x2


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      color = basestring, {
        'field': ('EnumField', (SimpleEnum,), {'default': SimpleEnum.RED})}

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'color')
    assert message_class.color.type is SimpleEnum
    assert message_class().color is SimpleEnum.RED

  def test_build_message_bogus_explicit_field(self):

    """ Test `build_message` with a bogus field value """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      color = basestring, {'field': 5.5}

    with self.assertRaises(TypeError):
      protorpc.build_message(SimpleModel)

  def test_build_message_submodel(self):

    """ Test `build_message` with an embedded submodel """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int


    class SimpleContainer(model.Model):

      """ Simple container message """

      model = SimpleModel

    message_class = protorpc.build_message(SimpleContainer)

    assert hasattr(message_class, 'model')
    assert message_class.__name__ == SimpleContainer.kind()
    assert message_class.model.message_type.__name__ == SimpleModel.kind()

  def test_build_message_variant(self):

    """ Test `build_message` with a variant property """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int
      data = dict

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert hasattr(message_class, 'data')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=5, data={'hi': 'sup'})
    message = message_class(string='hi', integer=5, data={'hi': 'sup'})

    assert _model.string == message.string
    assert _model.integer == message.integer
    assert _model.data['hi'] == message.data['hi']

  def test_build_message_variant_vanilla_model(self):

    """ Test `build_message` with a variant property (by vanilla `Model`) """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int
      data = model.Model

    some_model = SimpleModel(string='hithere')
    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert hasattr(message_class, 'data')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=5, data=some_model)
    message = message_class(string='hi', integer=5, data=some_model.to_dict())

    assert _model.string == message.string
    assert _model.integer == message.integer
    assert _model.data.string == message.data['string']

  def test_build_message_key(self):

    """ Test `build_message` with a `Key` property """


    class SimpleModel(model.Model):

      """ Simple model message """

      string = basestring
      integer = int
      ref = model.Key

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert hasattr(message_class, 'integer')
    assert hasattr(message_class, 'ref')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='hi', integer=5, ref=model.Key('hi', 1))
    message = message_class(string='hi', integer=5, ref=model.Key('hi', 1).to_message())
    _msg_from_obj = _model.to_message()

    assert _model.string == message.string
    assert _model.integer == message.integer
    assert _model.ref.urlsafe() == message.ref.encoded
    assert _msg_from_obj.ref.encoded == _model.ref.urlsafe()
    assert _msg_from_obj.ref.id == 1

  def test_build_message_hook(self):

    """ Test `build_message` with a target object hook """

    class SomeProperty(model.Property):

      """ Simple property with __message__ hook """

      _basetype = basestring

      @classmethod
      def __message__(cls, *args, **kwargs):

        """ Return an IntegerField implementation... """

        return messages.IntegerField(*args, **kwargs)

    class SimpleModel(model.Model):

      """ Simple model message """

      string = SomeProperty

    message_class = protorpc.build_message(SimpleModel)

    assert hasattr(message_class, 'string')
    assert message_class.__name__ == SimpleModel.kind()

    _model = SimpleModel(string='5')  # it's a basestring...
    message = message_class(string=5)  # jk its an int! lol

    assert int(_model.string) == message.string


class ProtoRPCAdaptedKeyTest(test.FrameworkTest):

  """ Tests for the ProtoRPC `Key` mixin """

  def test_key_to_message(self):

    """ Test converting a `Key` to a message """

    key = model.Key('Hi', 5)
    message = key.to_message()

    assert message.id == key.id
    assert message.kind == key.kind
    assert message.encoded == key.urlsafe()
    assert isinstance(message, rpc.Key)

  def test_key_with_name_to_message(self):

    """ Test converting a `Key` with a parent to a message """

    key = model.Key('Hi', 'there')
    message = key.to_message()

    assert message.id == key.id
    assert message.kind == key.kind
    assert message.encoded == key.urlsafe()
    assert isinstance(message, rpc.Key)

  def test_key_with_parent_to_message(self):

    """ Test converting a `Key` with a parent to a message """

    key = model.Key('Hi', 'there', parent=model.Key('Sup', 'hey'))
    message = key.to_message()

    assert message.id == key.id
    assert message.kind == key.kind
    assert message.encoded == key.urlsafe()
    assert message.parent.id == key.parent.id
    assert message.parent.kind == key.parent.kind
    assert message.parent.encoded == key.parent.urlsafe()
    assert isinstance(message, rpc.Key)
    assert isinstance(message.parent, rpc.Key)

  def test_key_with_ancestry_to_message(self):

    """ Test converting a `Key` with two parent generations to a message """

    key = model.Key('Hi', 'there', parent=model.Key('Sup', 'hey', parent=(
      model.Key('Hola', 'senor')
    )))

    message = key.to_message()

    assert isinstance(message, rpc.Key)
    assert isinstance(message.parent, rpc.Key)
    assert isinstance(message.parent.parent, rpc.Key)

    assert message.id == key.id
    assert message.kind == key.kind
    assert message.encoded == key.urlsafe()
    assert message.parent.id == key.parent.id
    assert message.parent.kind == key.parent.kind
    assert message.parent.encoded == key.parent.urlsafe()
    assert message.parent.parent.id == key.parent.parent.id
    assert message.parent.parent.kind == key.parent.parent.kind
    assert message.parent.parent.encoded == key.parent.parent.urlsafe()

  def test_key_to_message_model(self):

    """ Test converting a `Key` class to a message """

    message_class = model.Key.to_message_model()
    assert message_class.__name__ == model.Key.__name__

    assert hasattr(message_class, 'id')
    assert hasattr(message_class, 'kind')
    assert hasattr(message_class, 'parent')
    assert hasattr(message_class, 'encoded')
    assert hasattr(message_class, 'namespace')

  def test_message_to_key(self):

    """ Test inflating a message to a `Key` instance """

    message_class = model.Key.to_message_model()
    message = message_class(kind='Hi', id=5)

    assert hasattr(message_class, 'id')
    assert hasattr(message_class, 'kind')
    assert hasattr(message_class, 'parent')
    assert hasattr(message_class, 'encoded')
    assert hasattr(message_class, 'namespace')

    assert message.id == 5
    assert message.kind == 'Hi'

    key = model.Key.from_message(message)
    assert key.id == 5
    assert key.kind == 'Hi'

  def test_message_to_key_with_name(self):

    """ Test inflating a message to a `Key` with a string ID """

    message_class = model.Key.to_message_model()
    message = message_class(kind='Hi', id='friend')

    assert hasattr(message_class, 'id')
    assert hasattr(message_class, 'kind')
    assert hasattr(message_class, 'parent')
    assert hasattr(message_class, 'encoded')
    assert hasattr(message_class, 'namespace')

    assert message.id == 'friend'
    assert message.kind == 'Hi'

    key = model.Key.from_message(message)
    assert key.id == 'friend'
    assert key.kind == 'Hi'

  def test_message_to_key_with_parent(self):

    """ Test inflating a message to a `Key` with a parent """

    message_class = model.Key.to_message_model()
    message = message_class(
      kind='Hi',
      id='friend',
      parent=model.Key('Sup', 'homie').to_message())

    assert hasattr(message_class, 'id')
    assert hasattr(message_class, 'kind')
    assert hasattr(message_class, 'parent')
    assert hasattr(message_class, 'encoded')
    assert hasattr(message_class, 'namespace')

    assert message.id == 'friend'
    assert message.kind == 'Hi'

    key = model.Key.from_message(message)
    assert key.id == 'friend'
    assert key.kind == 'Hi'
    assert key.parent.id == 'homie'
    assert key.parent.kind == 'Sup'

  def test_message_to_key_with_ancestry(self):

    """ Test inflating a message to a `Key` with deep ancestry """

    message_class = model.Key.to_message_model()
    message = message_class(
      kind='Hi',
      id='friend',
      parent=model.Key('Sup', 'homie', **{
        'parent': model.Key('Hola', 'senor')
      }).to_message()
    )

    assert hasattr(message_class, 'id')
    assert hasattr(message_class, 'kind')
    assert hasattr(message_class, 'parent')
    assert hasattr(message_class, 'encoded')
    assert hasattr(message_class, 'namespace')

    assert message.id == 'friend'
    assert message.kind == 'Hi'

    key = model.Key.from_message(message)
    assert key.id == 'friend'
    assert key.kind == 'Hi'
    assert key.parent.id == 'homie'
    assert key.parent.kind == 'Sup'
    assert key.parent.parent.id == 'senor'
    assert key.parent.parent.kind == 'Hola'



class ProtoRPCAdaptedModelTests(test.FrameworkTest):

  """ Tests for the ProtoRPC `Model` mixin """

  def test_model_to_message(self):

    """ Test converting a `Model` to a `Message` """


    class SampleModel(model.Model):

      """ sample model """

      string = basestring
      number = int
      dt = datetime.datetime

    now = datetime.datetime.now()
    s = SampleModel(string='hi', number=5, dt=now)
    m = s.to_message()

    assert m.string == 'hi'
    assert m.number == 5
    assert m.dt == now.isoformat()

    s2 = SampleModel(key=model.Key(SampleModel, 'sup'), string='hi', number=5, dt=now)
    m2 = s2.to_message()

    assert m2.string == 'hi'
    assert m2.number == 5
    assert m2.dt == now.isoformat()
    assert m2.key.encoded == s2.key.urlsafe()

  """
  def test_model_to_message_model(self):

    ''' Test converting a `Model` class to a `Message` class '''

    pass

  def test_message_to_model(self):

    ''' Test inflating a `Model` from a `Message` '''

    pass
  """



# -*- coding: utf-8 -*-

"""

  debug utils
  ~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import sys

# 3rd party / stdlib
try:
  # noinspection PyPackageRequirements
  import logbook as logging
except ImportError:
  import logging
  logging.basicConfig(stream=sys.stdout, level=10, format='%(message)s')


def Logger(name):  # pragma: no cover

  """  """

  logger = logging.getLogger(name)
  logger.setLevel(10)
  return logger



# -*- coding: utf-8 -*-

"""

  model tests
  ~~~~~~~~~~~

  tests canteen's data modelling layer.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  # stdlib
  import os

  # canteen tests
  from canteen.test import FrameworkTest


  ## ModelExportTests
  class ModelExportTests(FrameworkTest):

    """ Tests objects exported by `model`. """

    def test_concrete(self):

      """ Test that we can import concrete classes. """

      try:
        from canteen import model
        from canteen.model import Key
        from canteen.model import Model
        from canteen.model import Property
        from canteen.model import AbstractKey
        from canteen.model import AbstractModel

      except ImportError:  # pragma: no cover
        return self.fail("Failed to import concrete classes exported by Model.")

      else:
        self.assertTrue(Key)  # must export Key
        self.assertTrue(Model)  # must export Model
        self.assertTrue(Property)  # must export Property
        self.assertTrue(AbstractKey)  # must export AbstractKey
        self.assertTrue(AbstractModel)  # must export AbstractModel
        self.assertIsInstance(model, type(os))  # must be a module (lol)


  __all__ = (
      'test_graph',
      'test_key',
      'test_meta',
      'test_model',
      'test_query')



# -*- coding: utf-8 -*-

"""

  core hook tests
  ~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# testing
from canteen import test

# hook and runtime subsystems
from canteen.core import hooks


class HookResponderTests(test.FrameworkTest):

  """ Tests ``hooks.HookResponder``, which allows the
      framework or downstream developers to execute
      code at discrete, named points in Canteen's
      execution flow. """

  def test_construct(self):

    """ Test simple construction of a `HookResponder` """

    r = hooks.HookResponder('sample-event')
    assert 'sample-event' in r.__hooks__

    r = hooks.HookResponder('sample-event', 'sample-event-2')
    assert 'sample-event' in r.__hooks__
    assert 'sample-event-2' in r.__hooks__

    r = hooks.HookResponder('sample-event', 'sample-event-2', rollup=True)
    assert 'sample-event' in r.__hooks__
    assert 'sample-event-2' in r.__hooks__
    assert r.__argspec__.__rollup__ is True

    r = hooks.HookResponder('sample-event', 'sample-event-2', rollup=False,
                                                              notify=True)
    assert 'sample-event' in r.__hooks__
    assert 'sample-event-2' in r.__hooks__
    assert r.__argspec__.__rollup__ is False
    assert r.__argspec__.__notify__ is True

    return r

  def test_implied_argspec(self):

    """ Test binding `Context` with an implied argspec """

    def responder(item_one):

      """ implied responder, satisfied using inspection """

      # i don't do anything
      assert item_one == 5
      return item_one

    r = hooks.HookResponder('test_event')(responder)

    # try executing
    r('test_event', item_one=5)

  def test_bind_context(self):

    """ Test binding `Context` for a `HookResponder` """

    c = hooks.Context(('sample_item', 'sample_item2'))

    def sup(*args, **kwargs): return kwargs

    # bind context
    unwrapped = c(sup)
    assert 'sample_item' in c.__requested__
    assert 'sample_item2' in c.__requested__
    assert c.__rollup__ is True
    assert c.__notify__ is False
    assert callable(unwrapped)

    result = unwrapped('sample-event', sample_item=True, sample_item2=False,
                                                         yompin=True)
    assert 'sample_item' in result and result['sample_item'] is True
    assert 'sample_item2' in result and result['sample_item2'] is False

  def test_bind_invalid_context(self):

    """ Test invalid binding behavior for `Context` """

    c = hooks.Context(('sample_item', 'sample_item2', 'sample_item3'))

    def sup(*args, **kwargs): return kwargs

    # bind context
    unwrapped = c(sup)
    assert 'sample_item' in c.__requested__
    assert 'sample_item2' in c.__requested__
    assert c.__rollup__ is True
    assert c.__notify__ is False
    assert callable(unwrapped)

    with self.assertRaises(RuntimeError):
      unwrapped('sample-event', sample_item=True, sample_item2=False,
                                                  yompin=True)



# -*- coding: utf-8 -*-

"""

  core runtime
  ~~~~~~~~~~~~

  platform internals and logic to discover/load/inject.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

from __future__ import print_function

# stdlib
import os
import sys
import abc
import inspect
import importlib
import threading

# core API
from .meta import Proxy
from .injection import Bridge


## Globals
__runtime__ = threading.local()


class Runtime(object):

  """  """

  # == Public Properties == #

  routes = None  # compiled route map
  config = None  # application config
  bridge = None  # window into the injection pool
  application = None  # WSGI application callable or delegate

  # == Private Properties == #
  __hooks__ = {}  # mapped hook points and methods to call
  __owner__ = "Runtime"  # meta bucket owner name for subclasses
  __wrapped__ = None  # wrapped dispatch method calculated on first request
  __singleton__ = False  # many runtimes can exist, _so power_
  __metaclass__ = Proxy.Component  # this should be injectable
  __precedence__ = False  # marked if a specific runtime should win in selection
  precedence = property(lambda self: self.__precedence__)  # protect writes

  # == Abstract Properties == #
  @staticmethod
  def base_exception():

    """  """

    return False

  @classmethod
  def spawn(cls, app):

    """  """

    global __runtime__
    if not getattr(__runtime__, 'active', None):
      __runtime__.active = (cls.resolve() if cls is Runtime else cls)(app)
    return __runtime__.active

  @classmethod
  def resolve(cls):

    """  """

    # @TODO(sgammon): figure out how to prioritize/select a runtime
    _default, _preferred = None, []
    for child in cls.iter_children():
      if hasattr(child, '__default__') and child.__default__:
        _default = child
        continue
      _preferred.append(child)

    for item in _preferred:
      if item.__precedence__:
        return item  # usually uWSGI
    if _preferred:
      return _preferred[0]  # Werkzeug
    return _default  # WSGIref

  @classmethod
  def set_precedence(cls, status=False):

    """  """

    return setattr(cls, '__precendence__', status) or cls

  @classmethod
  def add_hook(cls, hook, context_and_func):

    """  """

    context, func = context_and_func
    assert isinstance(hook, basestring), "hook name must be a string"
    if hook not in cls.__hooks__: cls.__hooks__[hook] = []
    cls.__hooks__[hook].append((context, func))
    return cls

  @classmethod
  def get_hooks(cls, point):

    """  """

    if point in cls.__hooks__:
      for i in cls.__hooks__[point]:
        yield i
    raise StopIteration()

  @classmethod
  def execute_hooks(cls, points, *args, **kwargs):

    """  """

    if isinstance(points, basestring): points = (points,)
    for point in points:
      for context, hook in cls.get_hooks(point):
        # noinspection PyBroadException
        try:
          # run as classmethod
          if isinstance(hook, classmethod):
            hook.__func__(context, *args, **kwargs)

          # run as staticmethod
          elif isinstance(hook, staticmethod):
            hook.__func__(*args, **kwargs)

          else:

            # must have a singleton if we're running in object context
            if not (
              hasattr(context, '__singleton__') or not context.__singleton__):
              raise RuntimeError('Cannot execute hook method "%s"'
                                 ' without matching singleton context.' % hook)

            # resolve singleton by context name
            obj = Proxy.Component.singleton_map.get(context.__name__)
            if not obj: raise RuntimeError('No matching singleton'
                                           ' for hook method "%s".' % hook)

            # run in singleton context
            hook(point, obj, *args, **kwargs)

        except Exception:
          if __debug__: raise

    return

  def __init__(self, app):

    """  """

    self.application, self.bridge = (
      app,
      Bridge())

  def initialize(self):

    """  """

    self.execute_hooks('initialize', runtime=self)

  def configure(self, config):

    """  """

    self.config = config
    self.initialize()  # let subclasses initialize
    return self

  def serve(self, interface, port, bind_only=False):

    """  """

    server = self.bind(interface, port)

    if bind_only:
      return server

    try:
      server.serve_forever()
    except (KeyboardInterrupt, Exception):
      print("Exiting.")
      sys.exit(0)

  def bind_environ(self, environ):

    """  """

    from ..logic import http
    self.routes = http.HTTPSemantics.route_map.bind_to_environ(environ)

    return (
      http.HTTPSemantics,
      http.HTTPSemantics.new_request(environ),
      http.HTTPSemantics.new_response())

  def handshake(self, key, origin=None):

    """ WIP """

    raise NotImplementedError('Runtime "%s" does not support'
                              ' realtime dispatch semantics. ' % self)

  def send(self, payload, binary=False):

    """ WIP """

    raise NotImplementedError('Runtime "%s" does not support'
                              ' realtime dispatch semantics. ' % self)

  def send(self):

    """ WIP """

    raise NotImplementedError('Runtime "%s" does not support'
                              ' realtime dispatch semantics. ' % self)

  def receive(self):

    """ WIP """

    raise NotImplementedError('Runtime "%s" does not support'
                              ' realtime dispatch semantics. ' % self)

  def dispatch(self, environ, start_response):

    """ WIP """

    from ..base import handler as base_handler

    # setup hook context
    context = {
      'environ': environ,
      'start_response': start_response,
      'runtime': self}

    # call dispatch hooks
    self.execute_hooks('dispatch', **context)

    # resolve URL via bound routes
    http, request, response = (
      context['http'],
      context['request'],
      context['response']
    ) = self.bind_environ(environ)

    # call request hooks
    self.execute_hooks('request', **context)

    # match route
    endpoint, arguments = (
      context['endpoint'],
      context['arguments']
    ) = self.routes.match()

    # call match hooks
    self.execute_hooks('match', **context)

    # resolve endpoint
    handler = context['handler'] = http.resolve_route(endpoint)

    if not handler:  # `None` for handler means it didn't match

      # update context
      context.update({
        'code': 404,
        'error': True,
        'exception': None,
        'response': None})

      # dispatch error hook for 404
      self.execute_hooks(('error', 'complete'), **context)

      # noinspection PyCallByClass,PyTypeChecker
      http.error(404)

    # class-based pages/handlers
    if isinstance(handler, type) and issubclass(handler, base_handler.Handler):

      # initialize handler
      flow = context['handler'] = handler(**context)

      # call handler hooks
      self.execute_hooks('handler', **context)

      # dispatch time: INCEPTION.
      result, iterator = flow(arguments), None

      if isinstance(result, tuple) and len(result) == 2:
        iterator, result = result  # extract iterator and raw result

      elif isinstance(result, tuple) and len(result) == 4:

        status, headers, content_type, content = (
          context['status'],
          context['headers'],
          context['content_type'],
          context['content']
        ) = result  # unpack response

        _response = context['response'] = response.__class__(content, **{
          'status': status,
          'headers': headers,
          'mimetype': content_type})

        # call response hooks
        self.execute_hooks(('response', 'complete'), **context)
        return _response(environ, start_response)

      # unpack response
      status, headers, content_type, content = (
        context['status'],
        context['headers'],
        context['content_type'],
        context['content']
      ) = result.status, result.headers, result.content_type, result.response

      # call response hooks
      self.execute_hooks(('response', 'complete'), **context)

      # send start_response
      start_response(result.status, [(
        k.encode('utf-8').strip(), v.encode('utf-8').strip()
      ) for k, v in result.headers])

      # buffer and return (i guess) @TODO(sgammon): can we do this better?
      return iterator or result.response  # it's a werkzeug Response

    # delegated class-based handlers (for instance, other WSGI apps)
    elif isinstance(handler, type) or callable(handler):

      # make a neat little shim, containing our runtime
      def _foreign_runtime_bridge(status, headers):

        """  """

        # call response hooks
        context['status'], context['headers'], context['response'] = (
          status,
          headers,
          None)

        self.execute_hooks(('response', 'complete'), **context)
        return start_response(status, headers)

      # attach runtime, arguments and actual start_response to shim
      _foreign_runtime_bridge.runtime = self
      _foreign_runtime_bridge.arguments = arguments
      _foreign_runtime_bridge.start_response = start_response
      context['start_response'] = _foreign_runtime_bridge

      # call hooks, initialize foreign handler with replaced start_response
      self.execute_hooks('handler', **context)
      return handler(environ, _foreign_runtime_bridge)

    # is it a function, maybe?
    if inspect.isfunction(handler):

      # inject stuff into context
      for prop, val in (
        ('runtime', self),
        ('self', self.bridge),
        ('arguments', arguments),
        ('request', request),
        ('response', response),
        ('environ', environ),
        ('start_response', start_response),
        ('Response', response.__class__)):

        handler.__globals__[prop] = val  # inject all the things

      # call handler hooks
      self.execute_hooks('handler', **context)

      # call with arguments only
      result = context['response'] = handler(**arguments)

      if isinstance(result, response.__class__):

        # call response hooks
        context['headers'], context['content'] = (
          result.headers, result.response)

        self.execute_hooks(('response', 'complete'), **context)

        # it's a Response class - delegate to attached start_response
        return response(environ, start_response)

      # a tuple bound to a URL - static response
      elif isinstance(result, tuple):

        if len(result) == 2:  # it's (status_code, response)
          status, response = (
            context['status'],
            context['response'],
          ) = result

          headers = context['headers'] = [
            ('Content-Type', 'text/html; charset=utf-8')]

          # call response hooks
          self.execute_hooks(('response', 'complete'), **context)
          start_response(status, headers)
          return iter([response])

        if len(result) == 3:  # it's (status_code, headers, response)
          status, headers, response = (
            context['status'],
            context['headers'],
            context['response']
          ) = result

          if isinstance(headers, dict):
            headers = headers.items()
            if 'Content-Type' not in headers:
              headers['Content-Type'] = context['headers']['Content-Type'] = (
                'text/html; charset=utf-8')

          # call response hooks
          self.execute_hooks(('response', 'complete'), **context)
          start_response(status, headers)
          return iter([response])

      elif isinstance(result, basestring):

        status, headers = (
          context['status'],
          context['headers'],
          context['response']
        ) = '200 OK', [('Content-Type', 'text/html; charset=utf-8')]

        # call response hooks
        self.execute_hooks(('response', 'complete'), **context)
        start_response(status, headers)
        return iter([result])

    # could be a bound response
    if not callable(handler):
      if isinstance(handler, basestring):

        context['status'], context['headers'], context['response'] = (
          '200 OK', [('Content-Type', 'text/html; charset=utf-8')], result)

        # call response hooks
        self.execute_hooks(('response', 'complete'), **context)
        return iter([handler])  # it's a static response!

    raise RuntimeError('Unrecognized handler type: "%s".' % type(handler))

  def wrap(self, dispatch):

    """  """

    if not self.__wrapped__:

      # default: return dispatch directly
      _dispatch = dispatch

      # == development wrappers
      dev_config = getattr(self.config, 'app', {}).get('dev', {})

      # profiler support
      if 'profiler' in dev_config:
        profiler_cfg = dev_config['profiler']

        if profiler_cfg.get('enable', False):

          ## grab a profiler
          try:
            import cProfile as profile
          except ImportError:
            import profile

          ## calculate dump file path
          profile_path = profiler_cfg.get('dump_file', os.path.abspath(
            os.path.join(*(os.getcwd(), '.develop', 'app.profile'))))

          ## current profile
          pkwargs = profiler_cfg.get('profile_kwargs', {})
          _current_profile = profile.Profile(**pkwargs)

          ## handle flushing mechanics
          if profiler_cfg.get('on_request', True):

            def maybe_flush_profile():

              """  """

              _current_profile.dump_stats(profile_path)

          else:
            # @TODO(sgammon): cross-request profiling
            raise RuntimeError('Cross-request profiling'
                               ' is currently unsupported.')

          def _dispatch(*args, **kwargs):

            """ Wrapper to enable profiler support. """

            ## dispatch
            response = _current_profile.runcall(dispatch, *args, **kwargs)
            maybe_flush_profile()
            return response

      self.__wrapped__ = _dispatch  # cache locally

    return self.__wrapped__

  def bind(self, interface, port):

    """  """

    raise NotImplementedError

  def callback(self, start_response):

    """  """

    def responder(status, headers):

      """  """

      return start_response(status, headers)

    return responder

  def __call__(self, environ, start_response):

    """  """

    try:
      return self.wrap(self.dispatch)(environ, self.callback(start_response))

    except self.base_exception as exc:
      return exc(environ, start_response)  # it's an acceptable exception

    except Exception:
      raise  # just raise it k?


class Library(object):

  """  """

  name = None  # string name of the library
  strict = False  # whether to hard-fail on ImportError
  package = None  # reference to the actual library package/module
  exception = None  # captured ImportError or AttributeError exception, if any
  supported = None  # boolean indicating whether this lib is supported or not

  __owner__, __metaclass__ = "Library", Proxy.Component

  def __init__(self, package, strict=False):

    """  """

    if isinstance(package, basestring):
      self.name = package
    elif isinstance(package, type(abc)):
      self.name, self.package, self.supported = package.__name__, package, True
    self.strict = strict

  def load(self, *subpackages):

    """  """

    loaded = []
    for package in subpackages:
      loaded.append(importlib.import_module('.'.join((self.name, package))))
    if len(loaded) == 1:
      return loaded[0]  # special case: one package only (return it directly)
    return tuple(loaded)  # otherwise, return a tuple of loaded modules

  def __enter__(self):

    """  """

    if not self.package and (self.supported is None):
      try:
        self.package = importlib.import_module(self.name)
      except ImportError as e:
        self.supported, self.exception = False, e
        if self.strict:
          raise
      else:
        self.supported = True
    return (self, self.package)

  def __exit__(self, exception_cls, exception, traceback):

    """  """

    if exception:
      if self.strict: return False

    return True
